# -*- coding: utf-8 -*-
"""
    :codeauthor: Thomas Jackson <jacksontj.89@gmail.com>
"""
from __future__ import absolute_import, print_function, unicode_literals

import ctypes
import multiprocessing
import os
import threading
import time
from concurrent.futures.thread import ThreadPoolExecutor

import salt.config
import salt.exceptions
import salt.ext.tornado.gen
import salt.ext.tornado.ioloop
import salt.log.setup
import salt.transport.client
import salt.transport.server
import salt.utils.platform
import salt.utils.process
import zmq.eventloop.ioloop
from salt.ext import six
from salt.ext.six.moves import range
from salt.ext.tornado.testing import AsyncTestCase
from salt.transport.zeromq import AsyncReqMessageClientPool
from saltfactories.utils.ports import get_unused_localhost_port
from tests.support.helpers import flaky, not_runs_on, slowTest
from tests.support.mixins import AdaptedConfigurationTestCaseMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase, skipIf
from tests.unit.transport.mixins import (
    PubChannelMixin,
    ReqChannelMixin,
    run_loop_in_thread,
)

# support pyzmq 13.0.x, TODO: remove once we force people to 14.0.x
if not hasattr(zmq.eventloop.ioloop, "ZMQIOLoop"):
    zmq.eventloop.ioloop.ZMQIOLoop = zmq.eventloop.ioloop.IOLoop


class BaseZMQReqCase(TestCase, AdaptedConfigurationTestCaseMixin):
    """
    Test the req server/client pair
    """

    @classmethod
    def setUpClass(cls):
        if not hasattr(cls, "_handle_payload"):
            return
        ret_port = get_unused_localhost_port()
        publish_port = get_unused_localhost_port()
        tcp_master_pub_port = get_unused_localhost_port()
        tcp_master_pull_port = get_unused_localhost_port()
        tcp_master_publish_pull = get_unused_localhost_port()
        tcp_master_workers = get_unused_localhost_port()
        cls.master_config = cls.get_temp_config(
            "master",
            **{
                "transport": "zeromq",
                "auto_accept": True,
                "ret_port": ret_port,
                "publish_port": publish_port,
                "tcp_master_pub_port": tcp_master_pub_port,
                "tcp_master_pull_port": tcp_master_pull_port,
                "tcp_master_publish_pull": tcp_master_publish_pull,
                "tcp_master_workers": tcp_master_workers,
            }
        )

        cls.minion_config = cls.get_temp_config(
            "minion",
            **{
                "transport": "zeromq",
                "master_ip": "127.0.0.1",
                "master_port": ret_port,
                "auth_timeout": 5,
                "auth_tries": 1,
                "master_uri": "tcp://127.0.0.1:{0}".format(ret_port),
            }
        )

        cls.process_manager = salt.utils.process.ProcessManager(
            name="ReqServer_ProcessManager"
        )

        cls.server_channel = salt.transport.server.ReqServerChannel.factory(
            cls.master_config
        )
        cls.server_channel.pre_fork(cls.process_manager)

        cls.io_loop = salt.ext.tornado.ioloop.IOLoop()
        cls.evt = threading.Event()
        cls.server_channel.post_fork(cls._handle_payload, io_loop=cls.io_loop)
        cls.server_thread = threading.Thread(
            target=run_loop_in_thread, args=(cls.io_loop, cls.evt)
        )
        cls.server_thread.start()

    @classmethod
    def tearDownClass(cls):
        if not hasattr(cls, "_handle_payload"):
            return
        # Attempting to kill the children hangs the test suite.
        # Let the test suite handle this instead.
        cls.process_manager.stop_restarting()
        cls.process_manager.kill_children()
        cls.evt.set()
        cls.server_thread.join()
        time.sleep(
            2
        )  # Give the procs a chance to fully close before we stop the io_loop
        cls.server_channel.close()
        del cls.server_channel
        del cls.io_loop
        del cls.process_manager
        del cls.server_thread
        del cls.master_config
        del cls.minion_config

    @classmethod
    def _handle_payload(cls, payload):
        """
        TODO: something besides echo
        """
        return payload, {"fun": "send_clear"}


class ClearReqTestCases(BaseZMQReqCase, ReqChannelMixin):
    """
    Test all of the clear msg stuff
    """

    def setUp(self):
        self.channel = salt.transport.client.ReqChannel.factory(
            self.minion_config, crypt="clear"
        )

    def tearDown(self):
        self.channel.close()
        del self.channel

    @classmethod
    @salt.ext.tornado.gen.coroutine
    def _handle_payload(cls, payload):
        """
        TODO: something besides echo
        """
        raise salt.ext.tornado.gen.Return((payload, {"fun": "send_clear"}))

    @slowTest
    def test_master_uri_override(self):
        """
        ensure master_uri kwarg is respected
        """
        # minion_config should be 127.0.0.1, we want a different uri that still connects
        uri = "tcp://{master_ip}:{master_port}".format(
            master_ip="localhost", master_port=self.minion_config["master_port"]
        )

        channel = salt.transport.client.ReqChannel.factory(
            self.minion_config, master_uri=uri
        )
        self.assertIn("localhost", channel.master_uri)
        del channel


@flaky
@not_runs_on(
    kernel="linux",
    os_familiy="Suse",
    reason="Skipping until https://github.com/saltstack/salt/issues/32902 gets fixed",
)
class AESReqTestCases(BaseZMQReqCase, ReqChannelMixin):
    def setUp(self):
        self.channel = salt.transport.client.ReqChannel.factory(self.minion_config)

    def tearDown(self):
        self.channel.close()
        del self.channel

    @classmethod
    @salt.ext.tornado.gen.coroutine
    def _handle_payload(cls, payload):
        """
        TODO: something besides echo
        """
        raise salt.ext.tornado.gen.Return((payload, {"fun": "send"}))

    # TODO: make failed returns have a specific framing so we can raise the same exception
    # on encrypted channels
    #
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    #
    # WARNING: This test will fail randomly on any system with > 1 CPU core!!!
    #
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    @slowTest
    def test_badload(self):
        """
        Test a variety of bad requests, make sure that we get some sort of error
        """
        # TODO: This test should be re-enabled when Jenkins moves to C7.
        # Once the version of salt-testing is increased to something newer than the September
        # release of salt-testing, the @flaky decorator should be applied to this test.
        msgs = ["", [], tuple()]
        for msg in msgs:
            with self.assertRaises(salt.exceptions.AuthenticationError):
                ret = self.channel.send(msg, timeout=5)


class BaseZMQPubCase(AsyncTestCase, AdaptedConfigurationTestCaseMixin):
    """
    Test the req server/client pair
    """

    @classmethod
    def setUpClass(cls):
        ret_port = get_unused_localhost_port()
        publish_port = get_unused_localhost_port()
        tcp_master_pub_port = get_unused_localhost_port()
        tcp_master_pull_port = get_unused_localhost_port()
        tcp_master_publish_pull = get_unused_localhost_port()
        tcp_master_workers = get_unused_localhost_port()
        cls.master_config = cls.get_temp_config(
            "master",
            **{
                "transport": "zeromq",
                "auto_accept": True,
                "ret_port": ret_port,
                "publish_port": publish_port,
                "tcp_master_pub_port": tcp_master_pub_port,
                "tcp_master_pull_port": tcp_master_pull_port,
                "tcp_master_publish_pull": tcp_master_publish_pull,
                "tcp_master_workers": tcp_master_workers,
            }
        )

        cls.minion_config = salt.config.minion_config(
            os.path.join(RUNTIME_VARS.TMP_CONF_DIR, "minion")
        )
        cls.minion_config = cls.get_temp_config(
            "minion",
            **{
                "transport": "zeromq",
                "master_ip": "127.0.0.1",
                "master_port": ret_port,
                "master_uri": "tcp://127.0.0.1:{0}".format(ret_port),
            }
        )

        cls.process_manager = salt.utils.process.ProcessManager(
            name="ReqServer_ProcessManager"
        )

        cls.server_channel = salt.transport.server.PubServerChannel.factory(
            cls.master_config
        )
        cls.server_channel.pre_fork(cls.process_manager)

        # we also require req server for auth
        cls.req_server_channel = salt.transport.server.ReqServerChannel.factory(
            cls.master_config
        )
        cls.req_server_channel.pre_fork(cls.process_manager)

        cls._server_io_loop = salt.ext.tornado.ioloop.IOLoop()
        cls.evt = threading.Event()
        cls.req_server_channel.post_fork(
            cls._handle_payload, io_loop=cls._server_io_loop
        )
        cls.server_thread = threading.Thread(
            target=run_loop_in_thread, args=(cls._server_io_loop, cls.evt)
        )
        cls.server_thread.start()

    @classmethod
    def tearDownClass(cls):
        cls.process_manager.kill_children()
        cls.process_manager.stop_restarting()
        time.sleep(
            2
        )  # Give the procs a chance to fully close before we stop the io_loop
        cls.evt.set()
        cls.server_thread.join()
        cls.req_server_channel.close()
        cls.server_channel.close()
        cls._server_io_loop.stop()
        del cls.server_channel
        del cls._server_io_loop
        del cls.process_manager
        del cls.server_thread
        del cls.master_config
        del cls.minion_config

    @classmethod
    def _handle_payload(cls, payload):
        """
        TODO: something besides echo
        """
        return payload, {"fun": "send_clear"}

    def setUp(self):
        super(BaseZMQPubCase, self).setUp()
        self._start_handlers = dict(self.io_loop._handlers)

    def tearDown(self):
        super(BaseZMQPubCase, self).tearDown()
        failures = []
        for k, v in six.iteritems(self.io_loop._handlers):
            if self._start_handlers.get(k) != v:
                failures.append((k, v))
        del self._start_handlers
        if len(failures) > 0:
            raise Exception("FDs still attached to the IOLoop: {0}".format(failures))


@skipIf(True, "Skip until we can devote time to fix this test")
class AsyncPubChannelTest(BaseZMQPubCase, PubChannelMixin):
    """
    Tests around the publish system
    """

    def get_new_ioloop(self):
        return salt.ext.tornado.ioloop.IOLoop()


class AsyncReqMessageClientPoolTest(TestCase):
    def setUp(self):
        super(AsyncReqMessageClientPoolTest, self).setUp()
        sock_pool_size = 5
        with patch(
            "salt.transport.zeromq.AsyncReqMessageClient.__init__",
            MagicMock(return_value=None),
        ):
            self.message_client_pool = AsyncReqMessageClientPool(
                {"sock_pool_size": sock_pool_size}, args=({}, "")
            )
        self.original_message_clients = self.message_client_pool.message_clients
        self.message_client_pool.message_clients = [
            MagicMock() for _ in range(sock_pool_size)
        ]

    def tearDown(self):
        del self.original_message_clients
        super(AsyncReqMessageClientPoolTest, self).tearDown()

    def test_send(self):
        for message_client_mock in self.message_client_pool.message_clients:
            message_client_mock.send_queue = [0, 0, 0]
            message_client_mock.send.return_value = []

        self.assertEqual([], self.message_client_pool.send())

        self.message_client_pool.message_clients[2].send_queue = [0]
        self.message_client_pool.message_clients[2].send.return_value = [1]
        self.assertEqual([1], self.message_client_pool.send())


class ZMQConfigTest(TestCase):
    def test_master_uri(self):
        """
        test _get_master_uri method
        """

        m_ip = "127.0.0.1"
        m_port = 4505
        s_ip = "111.1.0.1"
        s_port = 4058

        m_ip6 = "1234:5678::9abc"
        s_ip6 = "1234:5678::1:9abc"

        with patch("salt.transport.zeromq.LIBZMQ_VERSION_INFO", (4, 1, 6)), patch(
            "salt.transport.zeromq.ZMQ_VERSION_INFO", (16, 0, 1)
        ):
            # pass in both source_ip and source_port
            assert salt.transport.zeromq._get_master_uri(
                master_ip=m_ip, master_port=m_port, source_ip=s_ip, source_port=s_port
            ) == "tcp://{0}:{1};{2}:{3}".format(s_ip, s_port, m_ip, m_port)

            assert salt.transport.zeromq._get_master_uri(
                master_ip=m_ip6, master_port=m_port, source_ip=s_ip6, source_port=s_port
            ) == "tcp://[{0}]:{1};[{2}]:{3}".format(s_ip6, s_port, m_ip6, m_port)

            # source ip and source_port empty
            assert salt.transport.zeromq._get_master_uri(
                master_ip=m_ip, master_port=m_port
            ) == "tcp://{0}:{1}".format(m_ip, m_port)

            assert salt.transport.zeromq._get_master_uri(
                master_ip=m_ip6, master_port=m_port
            ) == "tcp://[{0}]:{1}".format(m_ip6, m_port)

            # pass in only source_ip
            assert salt.transport.zeromq._get_master_uri(
                master_ip=m_ip, master_port=m_port, source_ip=s_ip
            ) == "tcp://{0}:0;{1}:{2}".format(s_ip, m_ip, m_port)

            assert salt.transport.zeromq._get_master_uri(
                master_ip=m_ip6, master_port=m_port, source_ip=s_ip6
            ) == "tcp://[{0}]:0;[{1}]:{2}".format(s_ip6, m_ip6, m_port)

            # pass in only source_port
            assert salt.transport.zeromq._get_master_uri(
                master_ip=m_ip, master_port=m_port, source_port=s_port
            ) == "tcp://0.0.0.0:{0};{1}:{2}".format(s_port, m_ip, m_port)


class PubServerChannel(TestCase, AdaptedConfigurationTestCaseMixin):
    @classmethod
    def setUpClass(cls):
        ret_port = get_unused_localhost_port()
        publish_port = get_unused_localhost_port()
        tcp_master_pub_port = get_unused_localhost_port()
        tcp_master_pull_port = get_unused_localhost_port()
        tcp_master_publish_pull = get_unused_localhost_port()
        tcp_master_workers = get_unused_localhost_port()
        cls.master_config = cls.get_temp_config(
            "master",
            **{
                "transport": "zeromq",
                "auto_accept": True,
                "ret_port": ret_port,
                "publish_port": publish_port,
                "tcp_master_pub_port": tcp_master_pub_port,
                "tcp_master_pull_port": tcp_master_pull_port,
                "tcp_master_publish_pull": tcp_master_publish_pull,
                "tcp_master_workers": tcp_master_workers,
                "sign_pub_messages": False,
            }
        )
        salt.master.SMaster.secrets["aes"] = {
            "secret": multiprocessing.Array(
                ctypes.c_char, six.b(salt.crypt.Crypticle.generate_key_string()),
            ),
        }
        cls.minion_config = cls.get_temp_config(
            "minion",
            **{
                "transport": "zeromq",
                "master_ip": "127.0.0.1",
                "master_port": ret_port,
                "auth_timeout": 5,
                "auth_tries": 1,
                "master_uri": "tcp://127.0.0.1:{0}".format(ret_port),
            }
        )

    @classmethod
    def tearDownClass(cls):
        del cls.minion_config
        del cls.master_config

    def setUp(self):
        # Start the event loop, even though we don't directly use this with
        # ZeroMQPubServerChannel, having it running seems to increase the
        # likely hood of dropped messages.
        self.io_loop = salt.ext.tornado.ioloop.IOLoop()
        self.io_loop.make_current()
        self.io_loop_thread = threading.Thread(target=self.io_loop.start)
        self.io_loop_thread.start()
        self.process_manager = salt.utils.process.ProcessManager(
            name="PubServer_ProcessManager"
        )

    def tearDown(self):
        self.io_loop.add_callback(self.io_loop.stop)
        self.io_loop_thread.join()
        self.process_manager.stop_restarting()
        self.process_manager.kill_children()
        del self.io_loop
        del self.io_loop_thread
        del self.process_manager

    @staticmethod
    def _gather_results(opts, pub_uri, results, timeout=120, messages=None):
        """
        Gather results until then number of seconds specified by timeout passes
        without reveiving a message
        """
        ctx = zmq.Context()
        sock = ctx.socket(zmq.SUB)
        sock.setsockopt(zmq.LINGER, -1)
        sock.setsockopt(zmq.SUBSCRIBE, b"")
        sock.connect(pub_uri)
        last_msg = time.time()
        serial = salt.payload.Serial(opts)
        crypticle = salt.crypt.Crypticle(
            opts, salt.master.SMaster.secrets["aes"]["secret"].value
        )
        while time.time() - last_msg < timeout:
            try:
                payload = sock.recv(zmq.NOBLOCK)
            except zmq.ZMQError:
                time.sleep(0.01)
            else:
                if messages:
                    if messages != 1:
                        messages -= 1
                        continue
                payload = crypticle.loads(serial.loads(payload)["load"])
                if "stop" in payload:
                    break
                last_msg = time.time()
                results.append(payload["jid"])

    @skipIf(salt.utils.platform.is_windows(), "Skip on Windows OS")
    @slowTest
    def test_publish_to_pubserv_ipc(self):
        """
        Test sending 10K messags to ZeroMQPubServerChannel using IPC transport

        ZMQ's ipc transport not supported on Windows
        """
        opts = dict(self.master_config, ipc_mode="ipc", pub_hwm=0)
        server_channel = salt.transport.zeromq.ZeroMQPubServerChannel(opts)
        server_channel.pre_fork(
            self.process_manager,
            kwargs={"log_queue": salt.log.setup.get_multiprocessing_logging_queue()},
        )
        pub_uri = "tcp://{interface}:{publish_port}".format(**server_channel.opts)
        send_num = 10000
        expect = []
        results = []
        gather = threading.Thread(
            target=self._gather_results, args=(self.minion_config, pub_uri, results,)
        )
        gather.start()
        # Allow time for server channel to start, especially on windows
        time.sleep(2)
        for i in range(send_num):
            expect.append(i)
            load = {"tgt_type": "glob", "tgt": "*", "jid": i}
            server_channel.publish(load)
        server_channel.publish({"tgt_type": "glob", "tgt": "*", "stop": True})
        gather.join()
        server_channel.pub_close()
        assert len(results) == send_num, (len(results), set(expect).difference(results))

    @slowTest
    def test_zeromq_publish_port(self):
        """
        test when connecting that we
        use the publish_port set in opts
        when its not 4506
        """
        opts = dict(
            self.master_config,
            ipc_mode="ipc",
            pub_hwm=0,
            recon_randomize=False,
            publish_port=455505,
            recon_default=1,
            recon_max=2,
            master_ip="127.0.0.1",
            acceptance_wait_time=5,
            acceptance_wait_time_max=5,
        )
        opts["master_uri"] = "tcp://{interface}:{publish_port}".format(**opts)

        channel = salt.transport.zeromq.AsyncZeroMQPubChannel(opts)
        patch_socket = MagicMock(return_value=True)
        patch_auth = MagicMock(return_value=True)
        with patch.object(channel, "_socket", patch_socket), patch.object(
            channel, "auth", patch_auth
        ):
            channel.connect()
        assert str(opts["publish_port"]) in patch_socket.mock_calls[0][1][0]

    def test_zeromq_zeromq_filtering_decode_message_no_match(self):
        """
        test AsyncZeroMQPubChannel _decode_messages when
        zmq_filtering enabled and minion does not match
        """
        message = [
            b"4f26aeafdb2367620a393c973eddbe8f8b846eb",
            b"\x82\xa3enc\xa3aes\xa4load\xda\x00`\xeeR\xcf"
            b"\x0eaI#V\x17if\xcf\xae\x05\xa7\xb3bN\xf7\xb2\xe2"
            b'\xd0sF\xd1\xd4\xecB\xe8\xaf"/*ml\x80Q3\xdb\xaexg'
            b"\x8e\x8a\x8c\xd3l\x03\\,J\xa7\x01i\xd1:]\xe3\x8d"
            b"\xf4\x03\x88K\x84\n`\xe8\x9a\xad\xad\xc6\x8ea\x15>"
            b"\x92m\x9e\xc7aM\x11?\x18;\xbd\x04c\x07\x85\x99\xa3\xea[\x00D",
        ]

        opts = dict(
            self.master_config,
            ipc_mode="ipc",
            pub_hwm=0,
            zmq_filtering=True,
            recon_randomize=False,
            recon_default=1,
            recon_max=2,
            master_ip="127.0.0.1",
            acceptance_wait_time=5,
            acceptance_wait_time_max=5,
        )
        opts["master_uri"] = "tcp://{interface}:{publish_port}".format(**opts)

        server_channel = salt.transport.zeromq.AsyncZeroMQPubChannel(opts)
        with patch(
            "salt.crypt.AsyncAuth.crypticle",
            MagicMock(return_value={"tgt_type": "glob", "tgt": "*", "jid": 1}),
        ) as mock_test:
            res = server_channel._decode_messages(message)
        assert res.result() is None

    def test_zeromq_zeromq_filtering_decode_message(self):
        """
        test AsyncZeroMQPubChannel _decode_messages
        when zmq_filtered enabled
        """
        message = [
            b"4f26aeafdb2367620a393c973eddbe8f8b846ebd",
            b"\x82\xa3enc\xa3aes\xa4load\xda\x00`\xeeR\xcf"
            b"\x0eaI#V\x17if\xcf\xae\x05\xa7\xb3bN\xf7\xb2\xe2"
            b'\xd0sF\xd1\xd4\xecB\xe8\xaf"/*ml\x80Q3\xdb\xaexg'
            b"\x8e\x8a\x8c\xd3l\x03\\,J\xa7\x01i\xd1:]\xe3\x8d"
            b"\xf4\x03\x88K\x84\n`\xe8\x9a\xad\xad\xc6\x8ea\x15>"
            b"\x92m\x9e\xc7aM\x11?\x18;\xbd\x04c\x07\x85\x99\xa3\xea[\x00D",
        ]

        opts = dict(
            self.master_config,
            ipc_mode="ipc",
            pub_hwm=0,
            zmq_filtering=True,
            recon_randomize=False,
            recon_default=1,
            recon_max=2,
            master_ip="127.0.0.1",
            acceptance_wait_time=5,
            acceptance_wait_time_max=5,
        )
        opts["master_uri"] = "tcp://{interface}:{publish_port}".format(**opts)

        server_channel = salt.transport.zeromq.AsyncZeroMQPubChannel(opts)
        with patch(
            "salt.crypt.AsyncAuth.crypticle",
            MagicMock(return_value={"tgt_type": "glob", "tgt": "*", "jid": 1}),
        ) as mock_test:
            res = server_channel._decode_messages(message)

        assert res.result()["enc"] == "aes"

    @skipIf(salt.utils.platform.is_windows(), "Skip on Windows OS")
    @slowTest
    def test_zeromq_filtering(self):
        """
        Test sending messags to publisher using UDP
        with zeromq_filtering enabled
        """
        opts = dict(
            self.master_config,
            ipc_mode="ipc",
            pub_hwm=0,
            zmq_filtering=True,
            acceptance_wait_time=5,
        )
        server_channel = salt.transport.zeromq.ZeroMQPubServerChannel(opts)
        server_channel.pre_fork(
            self.process_manager,
            kwargs={"log_queue": salt.log.setup.get_multiprocessing_logging_queue()},
        )
        pub_uri = "tcp://{interface}:{publish_port}".format(**server_channel.opts)
        send_num = 1
        expect = []
        results = []
        gather = threading.Thread(
            target=self._gather_results,
            args=(self.minion_config, pub_uri, results,),
            kwargs={"messages": 2},
        )
        gather.start()
        # Allow time for server channel to start, especially on windows
        time.sleep(2)
        expect.append(send_num)
        load = {"tgt_type": "glob", "tgt": "*", "jid": send_num}
        with patch(
            "salt.utils.minions.CkMinions.check_minions",
            MagicMock(
                return_value={
                    "minions": ["minion"],
                    "missing": [],
                    "ssh_minions": False,
                }
            ),
        ):
            server_channel.publish(load)
        server_channel.publish({"tgt_type": "glob", "tgt": "*", "stop": True})
        gather.join()
        server_channel.pub_close()
        assert len(results) == send_num, (len(results), set(expect).difference(results))

    @slowTest
    def test_publish_to_pubserv_tcp(self):
        """
        Test sending 10K messags to ZeroMQPubServerChannel using TCP transport
        """
        opts = dict(self.master_config, ipc_mode="tcp", pub_hwm=0)
        server_channel = salt.transport.zeromq.ZeroMQPubServerChannel(opts)
        server_channel.pre_fork(
            self.process_manager,
            kwargs={"log_queue": salt.log.setup.get_multiprocessing_logging_queue()},
        )
        pub_uri = "tcp://{interface}:{publish_port}".format(**server_channel.opts)
        send_num = 10000
        expect = []
        results = []
        gather = threading.Thread(
            target=self._gather_results, args=(self.minion_config, pub_uri, results,)
        )
        gather.start()
        # Allow time for server channel to start, especially on windows
        time.sleep(2)
        for i in range(send_num):
            expect.append(i)
            load = {"tgt_type": "glob", "tgt": "*", "jid": i}
            server_channel.publish(load)
        gather.join()
        server_channel.pub_close()
        assert len(results) == send_num, (len(results), set(expect).difference(results))

    @staticmethod
    def _send_small(opts, sid, num=10):
        server_channel = salt.transport.zeromq.ZeroMQPubServerChannel(opts)
        for i in range(num):
            load = {"tgt_type": "glob", "tgt": "*", "jid": "{}-{}".format(sid, i)}
            server_channel.publish(load)
        server_channel.close()

    @staticmethod
    def _send_large(opts, sid, num=10, size=250000 * 3):
        server_channel = salt.transport.zeromq.ZeroMQPubServerChannel(opts)
        for i in range(num):
            load = {
                "tgt_type": "glob",
                "tgt": "*",
                "jid": "{}-{}".format(sid, i),
                "xdata": "0" * size,
            }
            server_channel.publish(load)
        server_channel.close()

    @slowTest
    def test_issue_36469_tcp(self):
        """
        Test sending both large and small messags to publisher using TCP

        https://github.com/saltstack/salt/issues/36469
        """
        opts = dict(self.master_config, ipc_mode="tcp", pub_hwm=0)
        server_channel = salt.transport.zeromq.ZeroMQPubServerChannel(opts)
        server_channel.pre_fork(
            self.process_manager,
            kwargs={"log_queue": salt.log.setup.get_multiprocessing_logging_queue()},
        )
        send_num = 10 * 4
        expect = []
        results = []
        pub_uri = "tcp://{interface}:{publish_port}".format(**opts)
        # Allow time for server channel to start, especially on windows
        time.sleep(2)
        gather = threading.Thread(
            target=self._gather_results, args=(self.minion_config, pub_uri, results,)
        )
        gather.start()
        with ThreadPoolExecutor(max_workers=4) as executor:
            executor.submit(self._send_small, opts, 1)
            executor.submit(self._send_small, opts, 2)
            executor.submit(self._send_small, opts, 3)
            executor.submit(self._send_large, opts, 4)
        expect = ["{}-{}".format(a, b) for a in range(10) for b in (1, 2, 3, 4)]
        time.sleep(0.1)
        server_channel.publish({"tgt_type": "glob", "tgt": "*", "stop": True})
        gather.join()
        server_channel.pub_close()
        assert len(results) == send_num, (len(results), set(expect).difference(results))



# -*- coding: utf-8 -*-
"""
Tests for win_path states
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

import copy

# Import Salt Libs
import salt.states.win_path as win_path

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, Mock, patch
from tests.support.unit import TestCase

NAME = "salt"


class WinPathTestCase(TestCase, LoaderModuleMockMixin):
    """
    Validate the win_path state
    """

    def setup_loader_modules(self):
        return {win_path: {}}

    def test_absent(self):
        """
        Test various cases for win_path.absent
        """
        ret_base = {"name": NAME, "result": True, "changes": {}}

        def _mock(retval):
            # Return a new MagicMock for each test case
            return MagicMock(side_effect=retval)

        # We don't really want to run the remove func
        with patch.dict(win_path.__salt__, {"win_path.remove": Mock()}):

            # Test mode OFF
            with patch.dict(win_path.__opts__, {"test": False}):

                # Test already absent
                with patch.dict(win_path.__salt__, {"win_path.exists": _mock([False])}):
                    ret = copy.deepcopy(ret_base)
                    ret["comment"] = "{0} is not in the PATH".format(NAME)
                    ret["result"] = True
                    self.assertDictEqual(win_path.absent(NAME), ret)

                # Test successful removal
                with patch.dict(
                    win_path.__salt__, {"win_path.exists": _mock([True, False])}
                ):
                    ret = copy.deepcopy(ret_base)
                    ret["comment"] = "Removed {0} from the PATH".format(NAME)
                    ret["changes"]["removed"] = NAME
                    ret["result"] = True
                    self.assertDictEqual(win_path.absent(NAME), ret)

                # Test unsucessful removal
                with patch.dict(
                    win_path.__salt__, {"win_path.exists": _mock([True, True])}
                ):
                    ret = copy.deepcopy(ret_base)
                    ret["comment"] = "Failed to remove {0} from the PATH".format(NAME)
                    ret["result"] = False
                    self.assertDictEqual(win_path.absent(NAME), ret)

            # Test mode ON
            with patch.dict(win_path.__opts__, {"test": True}):

                # Test already absent
                with patch.dict(win_path.__salt__, {"win_path.exists": _mock([False])}):
                    ret = copy.deepcopy(ret_base)
                    ret["comment"] = "{0} is not in the PATH".format(NAME)
                    ret["result"] = True
                    self.assertDictEqual(win_path.absent(NAME), ret)

                # Test the test-mode return
                with patch.dict(win_path.__salt__, {"win_path.exists": _mock([True])}):
                    ret = copy.deepcopy(ret_base)
                    ret["comment"] = "{0} would be removed from the PATH".format(NAME)
                    ret["result"] = None
                    self.assertDictEqual(win_path.absent(NAME), ret)

    def test_exists_invalid_index(self):
        """
        Tests win_path.exists when a non-integer index is specified.
        """
        ret = win_path.exists(NAME, index="foo")
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {},
                "result": False,
                "comment": "Index must be an integer",
            },
        )

    def test_exists_add_no_index_success(self):
        """
        Tests win_path.exists when the directory isn't already in the PATH and
        no index is specified (successful run).
        """
        add_mock = MagicMock(return_value=True)
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(
                side_effect=[["foo", "bar", "baz"], ["foo", "bar", "baz", NAME]]
            ),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": False}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME)

        add_mock.assert_called_once_with(NAME, index=None, rehash=False)
        self.assert_called_once(rehash_mock)
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {"index": {"old": None, "new": 3}},
                "result": True,
                "comment": "Added {0} to the PATH.".format(NAME),
            },
        )

    def test_exists_add_no_index_failure(self):
        """
        Tests win_path.exists when the directory isn't already in the PATH and
        no index is specified (failed run).
        """
        add_mock = MagicMock(return_value=False)
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(
                side_effect=[["foo", "bar", "baz"], ["foo", "bar", "baz"]]
            ),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": False}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME)

        add_mock.assert_called_once_with(NAME, index=None, rehash=False)
        rehash_mock.assert_not_called()
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {},
                "result": False,
                "comment": "Failed to add {0} to the PATH.".format(NAME),
            },
        )

    def test_exists_add_no_index_failure_exception(self):
        """
        Tests win_path.exists when the directory isn't already in the PATH and
        no index is specified (failed run due to exception).
        """
        add_mock = MagicMock(side_effect=Exception("Global Thermonuclear War"))
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(
                side_effect=[["foo", "bar", "baz"], ["foo", "bar", "baz"]]
            ),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": False}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME)

        add_mock.assert_called_once_with(NAME, index=None, rehash=False)
        rehash_mock.assert_not_called()
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {},
                "result": False,
                "comment": "Encountered error: Global Thermonuclear War. "
                "Failed to add {0} to the PATH.".format(NAME),
            },
        )

    def test_exists_change_index_success(self):
        """
        Tests win_path.exists when the directory is already in the PATH and
        needs to be moved to a different position (successful run).
        """
        add_mock = MagicMock(return_value=True)
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(
                side_effect=[["foo", "bar", "baz", NAME], [NAME, "foo", "bar", "baz"]]
            ),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": False}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME, index=0)

        add_mock.assert_called_once_with(NAME, index=0, rehash=False)
        self.assert_called_once(rehash_mock)
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {"index": {"old": 3, "new": 0}},
                "result": True,
                "comment": "Moved {0} from index 3 to 0.".format(NAME),
            },
        )

    def test_exists_change_negative_index_success(self):
        """
        Tests win_path.exists when the directory is already in the PATH and
        needs to be moved to a different position (successful run).

        This tests a negative index.
        """
        add_mock = MagicMock(return_value=True)
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(
                side_effect=[["foo", "bar", NAME, "baz"], ["foo", "bar", "baz", NAME]]
            ),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": False}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME, index=-1)

        add_mock.assert_called_once_with(NAME, index=-1, rehash=False)
        self.assert_called_once(rehash_mock)
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {"index": {"old": -2, "new": -1}},
                "result": True,
                "comment": "Moved {0} from index -2 to -1.".format(NAME),
            },
        )

    def test_exists_change_index_add_exception(self):
        """
        Tests win_path.exists when the directory is already in the PATH but an
        exception is raised when we attempt to add the key to its new location.
        """
        add_mock = MagicMock(side_effect=Exception("Global Thermonuclear War"))
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(
                side_effect=[["foo", "bar", "baz", NAME], ["foo", "bar", "baz", NAME]]
            ),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": False}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME, index=0)

        add_mock.assert_called_once_with(NAME, index=0, rehash=False)
        rehash_mock.assert_not_called()
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {},
                "result": False,
                "comment": "Encountered error: Global Thermonuclear War. "
                "Failed to move {0} from index 3 to 0.".format(NAME),
            },
        )

    def test_exists_change_negative_index_add_exception(self):
        """
        Tests win_path.exists when the directory is already in the PATH but an
        exception is raised when we attempt to add the key to its new location.

        This tests a negative index.
        """
        add_mock = MagicMock(side_effect=Exception("Global Thermonuclear War"))
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(
                side_effect=[["foo", "bar", NAME, "baz"], ["foo", "bar", NAME, "baz"]]
            ),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": False}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME, index=-1)

        add_mock.assert_called_once_with(NAME, index=-1, rehash=False)
        rehash_mock.assert_not_called()
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {},
                "result": False,
                "comment": "Encountered error: Global Thermonuclear War. "
                "Failed to move {0} from index -2 to -1.".format(NAME),
            },
        )

    def test_exists_change_index_failure(self):
        """
        Tests win_path.exists when the directory is already in the PATH and
        needs to be moved to a different position (failed run).
        """
        add_mock = MagicMock(return_value=False)
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(
                side_effect=[["foo", "bar", "baz", NAME], ["foo", "bar", "baz", NAME]]
            ),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": False}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME, index=0)

        add_mock.assert_called_once_with(NAME, index=0, rehash=False)
        rehash_mock.assert_not_called()
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {},
                "result": False,
                "comment": "Failed to move {0} from index 3 to 0.".format(NAME),
            },
        )

    def test_exists_change_negative_index_failure(self):
        """
        Tests win_path.exists when the directory is already in the PATH and
        needs to be moved to a different position (failed run).

        This tests a negative index.
        """
        add_mock = MagicMock(return_value=False)
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(
                side_effect=[["foo", "bar", NAME, "baz"], ["foo", "bar", NAME, "baz"]]
            ),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": False}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME, index=-1)

        add_mock.assert_called_once_with(NAME, index=-1, rehash=False)
        rehash_mock.assert_not_called()
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {},
                "result": False,
                "comment": "Failed to move {0} from index -2 to -1.".format(NAME),
            },
        )

    def test_exists_change_index_test_mode(self):
        """
        Tests win_path.exists when the directory is already in the PATH and
        needs to be moved to a different position (test mode enabled).
        """
        add_mock = Mock()
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(side_effect=[["foo", "bar", "baz", NAME]]),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": True}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME, index=0)

        add_mock.assert_not_called()
        rehash_mock.assert_not_called()
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {"index": {"old": 3, "new": 0}},
                "result": None,
                "comment": "{0} would be moved from index 3 to 0.".format(NAME),
            },
        )

    def test_exists_change_negative_index_test_mode(self):
        """
        Tests win_path.exists when the directory is already in the PATH and
        needs to be moved to a different position (test mode enabled).
        """
        add_mock = Mock()
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(side_effect=[["foo", "bar", NAME, "baz"]]),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": True}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME, index=-1)

        add_mock.assert_not_called()
        rehash_mock.assert_not_called()
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {"index": {"old": -2, "new": -1}},
                "result": None,
                "comment": "{0} would be moved from index -2 to -1.".format(NAME),
            },
        )

    def _test_exists_add_already_present(self, index, test_mode):
        """
        Tests win_path.exists when the directory already exists in the PATH.
        Helper function to test both with and without and index, and with test
        mode both disabled and enabled.
        """
        current_path = ["foo", "bar", "baz"]
        if index is None:
            current_path.append(NAME)
        else:
            pos = index if index >= 0 else len(current_path) + index + 1
            current_path.insert(pos, NAME)

        add_mock = Mock()
        rehash_mock = MagicMock(return_value=True)
        dunder_salt = {
            "win_path.get_path": MagicMock(side_effect=[current_path]),
            "win_path.add": add_mock,
            "win_path.rehash": rehash_mock,
        }
        dunder_opts = {"test": test_mode}

        with patch.dict(win_path.__salt__, dunder_salt), patch.dict(
            win_path.__opts__, dunder_opts
        ):
            ret = win_path.exists(NAME, index=index)

        add_mock.assert_not_called()
        rehash_mock.assert_not_called()
        self.assertDictEqual(
            ret,
            {
                "name": NAME,
                "changes": {},
                "result": True,
                "comment": "{0} already exists in the PATH{1}.".format(
                    NAME, " at index {0}".format(index) if index is not None else ""
                ),
            },
        )

    def test_exists_add_no_index_already_present(self):
        self._test_exists_add_already_present(None, False)

    def test_exists_add_no_index_already_present_test_mode(self):
        self._test_exists_add_already_present(None, True)

    def test_exists_add_index_already_present(self):
        self._test_exists_add_already_present(1, False)
        self._test_exists_add_already_present(2, False)
        self._test_exists_add_already_present(-1, False)
        self._test_exists_add_already_present(-2, False)

    def test_exists_add_index_already_present_test_mode(self):
        self._test_exists_add_already_present(1, True)
        self._test_exists_add_already_present(2, True)
        self._test_exists_add_already_present(-1, True)
        self._test_exists_add_already_present(-2, True)



# -*- coding: utf-8 -*-
"""
tests.unit.setup.test_install
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
from __future__ import absolute_import, print_function, unicode_literals

import json
import logging
import os
import pathlib
import re
import sys

import salt.utils.path
import salt.utils.platform
import salt.version
from salt.modules.virtualenv_mod import KNOWN_BINARY_NAMES
from tests.support.helpers import VirtualEnv, slowTest, with_tempdir
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase, skipIf

log = logging.getLogger(__name__)


@skipIf(
    salt.utils.path.which_bin(KNOWN_BINARY_NAMES) is None, "virtualenv not installed"
)
class InstallTest(TestCase):
    """
    Tests for building and installing salt
    """

    @slowTest
    @with_tempdir()
    def test_wheel(self, tempdir):
        """
        test building and installing a bdist_wheel package
        """
        # Let's create the testing virtualenv
        with VirtualEnv() as venv:
            venv.run(venv.venv_python, "setup.py", "clean", cwd=RUNTIME_VARS.CODE_DIR)
            venv.run(
                venv.venv_python,
                "setup.py",
                "bdist_wheel",
                "--dist-dir",
                tempdir,
                cwd=RUNTIME_VARS.CODE_DIR,
            )
            venv.run(venv.venv_python, "setup.py", "clean", cwd=RUNTIME_VARS.CODE_DIR)

            salt_generated_package = list(pathlib.Path(tempdir).glob("*.whl"))
            if not salt_generated_package:
                self.fail("Could not find the generated wheel file")
            salt_generated_package = salt_generated_package[0]

            # Assert generate wheel version matches what salt reports as its version
            whl_ver = [
                x
                for x in salt_generated_package.name.split("-")
                if re.search(r"^\d.\d*", x)
            ][0]
            whl_ver_cmp = whl_ver.replace("_", "-")
            salt_ver_cmp = salt.version.__version__.replace("/", "-")
            assert whl_ver_cmp == salt_ver_cmp, "{} != {}".format(
                whl_ver_cmp, salt_ver_cmp
            )

            # Because bdist_wheel supports pep517, we don't have to pre-install Salt's
            # dependencies before installing the wheel package
            venv.install(str(salt_generated_package))

            # Let's ensure the version is correct
            cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
            for details in json.loads(cmd.stdout):
                if details["name"] != "salt":
                    continue
                installed_version = details["version"]
                break
            else:
                self.fail("Salt was not found installed")

            # Let's compare the installed version with the version salt reports
            assert installed_version == salt_ver_cmp, "{} != {}".format(
                installed_version, salt_ver_cmp
            )

            # Let's also ensure we have a salt/_version.py from the installed salt wheel
            subdir = [
                "lib",
                "python{}.{}".format(*sys.version_info),
                "site-packages",
                "salt",
            ]
            if salt.utils.platform.is_windows():
                subdir.pop(1)

            installed_salt_path = pathlib.Path(venv.venv_dir)
            installed_salt_path = installed_salt_path.joinpath(*subdir)
            assert installed_salt_path.is_dir()
            salt_generated_version_file_path = installed_salt_path / "_version.py"
            assert salt_generated_version_file_path.is_file()

    @slowTest
    @with_tempdir()
    def test_egg(self, tempdir):
        """
        test building and installing a bdist_egg package
        """
        # TODO: We should actually dissallow generating an egg file
        # Let's create the testing virtualenv
        with VirtualEnv() as venv:
            venv.run(venv.venv_python, "setup.py", "clean", cwd=RUNTIME_VARS.CODE_DIR)

            # Setuptools installs pre-release packages if we don't pin to an exact version
            # Let's download and install requirements before, running salt's install test
            venv.run(
                venv.venv_python,
                "-m",
                "pip",
                "download",
                "--dest",
                tempdir,
                RUNTIME_VARS.CODE_DIR,
            )
            packages = []
            for fname in os.listdir(tempdir):
                packages.append(os.path.join(tempdir, fname))
            venv.install(*packages)
            for package in packages:
                os.unlink(package)

            venv.run(
                venv.venv_python,
                "setup.py",
                "bdist_egg",
                "--dist-dir",
                tempdir,
                cwd=RUNTIME_VARS.CODE_DIR,
            )
            venv.run(venv.venv_python, "setup.py", "clean", cwd=RUNTIME_VARS.CODE_DIR)

            salt_generated_package = list(pathlib.Path(tempdir).glob("*.egg"))
            if not salt_generated_package:
                self.fail("Could not find the generated egg file")
            salt_generated_package = salt_generated_package[0]

            # Assert generate wheel version matches what salt reports as its version
            egg_ver = [
                x
                for x in salt_generated_package.name.split("-")
                if re.search(r"^\d.\d*", x)
            ][0]
            egg_ver_cmp = egg_ver.replace("_", "-")
            salt_ver_cmp = salt.version.__version__.replace("/", "-")
            assert egg_ver_cmp == salt_ver_cmp, "{} != {}".format(
                egg_ver_cmp, salt_ver_cmp
            )

            # We cannot pip install an egg file, let's go old school
            venv.run(
                venv.venv_python, "-m", "easy_install", str(salt_generated_package)
            )

            # Let's ensure the version is correct
            cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
            for details in json.loads(cmd.stdout):
                if details["name"] != "salt":
                    continue
                installed_version = details["version"]
                break
            else:
                self.fail("Salt was not found installed")

            # Let's compare the installed version with the version salt reports
            assert installed_version == salt_ver_cmp, "{} != {}".format(
                installed_version, salt_ver_cmp
            )

            # Let's also ensure we have a salt/_version.py from the installed salt egg
            subdir = [
                "lib",
                "python{}.{}".format(*sys.version_info),
                "site-packages",
            ]
            if salt.utils.platform.is_windows():
                subdir.pop(1)
            site_packages_dir = pathlib.Path(venv.venv_dir)
            site_packages_dir = site_packages_dir.joinpath(*subdir)
            assert site_packages_dir.is_dir()
            installed_salt_path = list(site_packages_dir.glob("salt*.egg"))
            if not installed_salt_path:
                self.fail("Failed to find the installed salt path")
            log.debug("Installed salt path glob matches: %s", installed_salt_path)
            installed_salt_path = installed_salt_path[0] / "salt"
            assert installed_salt_path.is_dir()
            salt_generated_version_file_path = installed_salt_path / "_version.py"
            assert (
                salt_generated_version_file_path.is_file()
            ), "{} is not a file".format(salt_generated_version_file_path)

    # On python 3.5 Windows sdist fails with encoding errors. This is resolved
    # in later versions.
    @skipIf(
        salt.utils.platform.is_windows()
        and sys.version_info > (3,)
        and sys.version_info < (3, 6),
        "Skip on python 3.5",
    )
    @slowTest
    @with_tempdir()
    def test_sdist(self, tempdir):
        """
        test building and installing a sdist package
        """
        # Let's create the testing virtualenv
        with VirtualEnv() as venv:
            venv.run(venv.venv_python, "setup.py", "clean", cwd=RUNTIME_VARS.CODE_DIR)

            # Setuptools installs pre-release packages if we don't pin to an exact version
            # Let's download and install requirements before, running salt's install test
            venv.run(
                venv.venv_python,
                "-m",
                "pip",
                "download",
                "--dest",
                tempdir,
                RUNTIME_VARS.CODE_DIR,
            )
            packages = []
            for fname in os.listdir(tempdir):
                packages.append(os.path.join(tempdir, fname))
            venv.install(*packages)
            for package in packages:
                os.unlink(package)

            venv.run(
                venv.venv_python,
                "setup.py",
                "sdist",
                "--dist-dir",
                tempdir,
                cwd=RUNTIME_VARS.CODE_DIR,
            )
            venv.run(venv.venv_python, "setup.py", "clean", cwd=RUNTIME_VARS.CODE_DIR)

            salt_generated_package = list(pathlib.Path(tempdir).glob("*.tar.gz"))
            if not salt_generated_package:
                self.fail("Could not find the generated sdist file")
            salt_generated_package = salt_generated_package[0]
            log.info("Generated sdist file: %s", salt_generated_package.name)

            # Assert generated sdist version matches what salt reports as its version
            sdist_ver_cmp = salt_generated_package.name.split(".tar.gz")[0].split(
                "salt-"
            )[-1]
            salt_ver_cmp = salt.version.__version__.replace("/", "-")
            assert sdist_ver_cmp == salt_ver_cmp, "{} != {}".format(
                sdist_ver_cmp, salt.version.__version__
            )

            venv.install(str(salt_generated_package))

            # Let's also ensure we have a salt/_version.py from the installed salt wheel
            subdir = [
                "lib",
                "python{}.{}".format(*sys.version_info),
                "site-packages",
                "salt",
            ]
            if salt.utils.platform.is_windows():
                subdir.pop(1)

            installed_salt_path = pathlib.Path(venv.venv_dir)
            installed_salt_path = installed_salt_path.joinpath(*subdir)
            assert installed_salt_path.is_dir()
            salt_generated_version_file_path = installed_salt_path / "_version.py"
            assert salt_generated_version_file_path.is_file()
            with salt_generated_version_file_path.open() as rfh:
                log.debug("_version.py contents:\n >>>>>>\n%s\n <<<<<<", rfh.read())

            # Let's ensure the version is correct
            cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
            for details in json.loads(cmd.stdout):
                if details["name"] != "salt":
                    continue
                installed_version = details["version"]
                break
            else:
                self.fail("Salt was not found installed")

            # Let's compare the installed version with the version salt reports
            assert installed_version == salt_ver_cmp, "{} != {}".format(
                installed_version, salt_ver_cmp
            )

    @slowTest
    @with_tempdir()
    def test_setup_install(self, tempdir):
        """
        test installing directly from source
        """
        # Let's create the testing virtualenv
        with VirtualEnv() as venv:
            venv.run(venv.venv_python, "setup.py", "clean", cwd=RUNTIME_VARS.CODE_DIR)

            # Setuptools installs pre-release packages if we don't pin to an exact version
            # Let's download and install requirements before, running salt's install test
            venv.run(
                venv.venv_python,
                "-m",
                "pip",
                "download",
                "--dest",
                tempdir,
                RUNTIME_VARS.CODE_DIR,
            )
            packages = []
            for fname in os.listdir(tempdir):
                packages.append(os.path.join(tempdir, fname))
            venv.install(*packages)
            for package in packages:
                os.unlink(package)

            venv.run(
                venv.venv_python,
                "setup.py",
                "install",
                "--prefix",
                venv.venv_dir,
                cwd=RUNTIME_VARS.CODE_DIR,
            )

            venv.run(venv.venv_python, "setup.py", "clean", cwd=RUNTIME_VARS.CODE_DIR)

            # Let's ensure the version is correct
            cmd = venv.run(venv.venv_python, "-m", "pip", "list", "--format", "json")
            for details in json.loads(cmd.stdout):
                if details["name"] != "salt":
                    continue
                installed_version = details["version"]
                break
            else:
                self.fail("Salt was not found installed")

            salt_ver_cmp = salt.version.__version__.replace("/", "-")
            # Let's compare the installed version with the version salt reports
            assert installed_version == salt_ver_cmp, "{} != {}".format(
                installed_version, salt_ver_cmp
            )

            # Let's also ensure we have a salt/_version.py from the installed salt
            subdir = [
                "lib",
                "python{}.{}".format(*sys.version_info),
                "site-packages",
            ]
            if salt.utils.platform.is_windows():
                subdir.pop(1)
            site_packages_dir = pathlib.Path(venv.venv_dir)
            site_packages_dir = site_packages_dir.joinpath(*subdir)
            assert site_packages_dir.is_dir()
            installed_salt_path = list(site_packages_dir.glob("salt*.egg"))
            if not installed_salt_path:
                self.fail("Failed to find the installed salt path")
            installed_salt_path = installed_salt_path[0] / "salt"
            salt_generated_version_file_path = installed_salt_path / "_version.py"
            assert salt_generated_version_file_path.is_file()



# -*- coding: utf-8 -*-
"""
    :codeauthor: jmoney <justin@saltstack.com>
"""

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.modules.linux_sysctl as linux_sysctl
import salt.modules.systemd_service as systemd
from salt.exceptions import CommandExecutionError

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, mock_open, patch
from tests.support.unit import TestCase


class LinuxSysctlTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.modules.linux_sysctl module
    """

    def setup_loader_modules(self):
        return {linux_sysctl: {}, systemd: {}}

    def test_get(self):
        """
        Tests the return of get function
        """
        mock_cmd = MagicMock(return_value=1)
        with patch.dict(linux_sysctl.__salt__, {"cmd.run": mock_cmd}):
            self.assertEqual(linux_sysctl.get("net.ipv4.ip_forward"), 1)

    def test_get_ignore(self):
        """
        Tests the return of get function with ignore
        """
        mock_cmd = MagicMock(return_value="")
        with patch.dict(linux_sysctl.__salt__, {"cmd.run": mock_cmd}):
            self.assertEqual(linux_sysctl.get("net.ipv4.ip_forward", ignore=True), "")

    def test_assign_proc_sys_failed(self):
        """
        Tests if /proc/sys/<kernel-subsystem> exists or not
        """
        with patch("os.path.exists", MagicMock(return_value=False)):
            cmd = {
                "pid": 1337,
                "retcode": 0,
                "stderr": "",
                "stdout": "net.ipv4.ip_forward = 1",
            }
            mock_cmd = MagicMock(return_value=cmd)
            with patch.dict(linux_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
                self.assertRaises(
                    CommandExecutionError, linux_sysctl.assign, "net.ipv4.ip_forward", 1
                )

    def test_assign_cmd_failed(self):
        """
        Tests if the assignment was successful or not
        """
        with patch("os.path.exists", MagicMock(return_value=True)):
            cmd = {
                "pid": 1337,
                "retcode": 0,
                "stderr": 'sysctl: setting key "net.ipv4.ip_forward": Invalid argument',
                "stdout": "net.ipv4.ip_forward = backward",
            }
            mock_cmd = MagicMock(return_value=cmd)
            with patch.dict(linux_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
                self.assertRaises(
                    CommandExecutionError,
                    linux_sysctl.assign,
                    "net.ipv4.ip_forward",
                    "backward",
                )

    def test_assign_success(self):
        """
        Tests the return of successful assign function
        """
        with patch("os.path.exists", MagicMock(return_value=True)):
            cmd = {
                "pid": 1337,
                "retcode": 0,
                "stderr": "",
                "stdout": "net.ipv4.ip_forward = 1",
            }
            ret = {"net.ipv4.ip_forward": "1"}
            mock_cmd = MagicMock(return_value=cmd)
            with patch.dict(linux_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
                self.assertEqual(linux_sysctl.assign("net.ipv4.ip_forward", 1), ret)

    def test_assign_ignore(self):
        """
        Tests the ignore assign function
        """
        with patch("os.path.exists", MagicMock(return_value=True)):
            cmd = {"pid": 1337, "retcode": 0, "stderr": "", "stdout": ""}
            ret = {"net.ipv4.ip_forward": "ignored"}
            mock_cmd = MagicMock(return_value=cmd)
            with patch.dict(linux_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
                self.assertEqual(
                    linux_sysctl.assign("net.ipv4.ip_forward", 1, ignore=True), ret
                )

    def test_persist_no_conf_failure(self):
        """
        Tests adding of config file failure
        """
        asn_cmd = {
            "pid": 1337,
            "retcode": 0,
            "stderr": "sysctl: permission denied",
            "stdout": "",
        }
        mock_asn_cmd = MagicMock(return_value=asn_cmd)
        cmd = "sysctl -w net.ipv4.ip_forward=1"
        mock_cmd = MagicMock(return_value=cmd)
        with patch.dict(
            linux_sysctl.__salt__,
            {"cmd.run_stdout": mock_cmd, "cmd.run_all": mock_asn_cmd},
        ):
            with patch("salt.utils.files.fopen", mock_open()) as m_open:
                self.assertRaises(
                    CommandExecutionError,
                    linux_sysctl.persist,
                    "net.ipv4.ip_forward",
                    1,
                    config=None,
                )

    def test_persist_no_conf_success(self):
        """
        Tests successful add of config file when previously not one
        """
        config = "/etc/sysctl.conf"
        with patch("os.path.isfile", MagicMock(return_value=False)), patch(
            "os.path.exists", MagicMock(return_value=True)
        ):
            asn_cmd = {
                "pid": 1337,
                "retcode": 0,
                "stderr": "",
                "stdout": "net.ipv4.ip_forward = 1",
            }
            mock_asn_cmd = MagicMock(return_value=asn_cmd)

            sys_cmd = "systemd 208\n+PAM +LIBWRAP"
            mock_sys_cmd = MagicMock(return_value=sys_cmd)

            with patch("salt.utils.files.fopen", mock_open()) as m_open, patch.dict(
                linux_sysctl.__context__, {"salt.utils.systemd.version": 232}
            ), patch.dict(
                linux_sysctl.__salt__,
                {"cmd.run_stdout": mock_sys_cmd, "cmd.run_all": mock_asn_cmd},
            ), patch.dict(
                systemd.__context__,
                {"salt.utils.systemd.booted": True, "salt.utils.systemd.version": 232},
            ):
                linux_sysctl.persist("net.ipv4.ip_forward", 1, config=config)
                writes = m_open.write_calls()
                assert writes == ["#\n# Kernel sysctl configuration\n#\n"], writes

    def test_persist_read_conf_success(self):
        """
        Tests sysctl.conf read success
        """
        with patch("os.path.isfile", MagicMock(return_value=True)), patch(
            "os.path.exists", MagicMock(return_value=True)
        ):
            asn_cmd = {
                "pid": 1337,
                "retcode": 0,
                "stderr": "",
                "stdout": "net.ipv4.ip_forward = 1",
            }
            mock_asn_cmd = MagicMock(return_value=asn_cmd)

            sys_cmd = "systemd 208\n+PAM +LIBWRAP"
            mock_sys_cmd = MagicMock(return_value=sys_cmd)

            with patch("salt.utils.files.fopen", mock_open()):
                with patch.dict(
                    linux_sysctl.__context__, {"salt.utils.systemd.version": 232}
                ):
                    with patch.dict(
                        linux_sysctl.__salt__,
                        {"cmd.run_stdout": mock_sys_cmd, "cmd.run_all": mock_asn_cmd},
                    ):
                        with patch.dict(
                            systemd.__context__, {"salt.utils.systemd.booted": True}
                        ):
                            self.assertEqual(
                                linux_sysctl.persist("net.ipv4.ip_forward", 1),
                                "Updated",
                            )



# -*- coding: utf-8 -*-
"""
Primary interfaces for the salt-cloud system
"""
# Need to get data from 4 sources!
# CLI options
# salt cloud config - CONFIG_DIR + '/cloud'
# salt master config (for master integration)
# salt VM config, where VMs are defined - CONFIG_DIR + '/cloud.profiles'
#
# The cli, master and cloud configs will merge for opts
# the VM data will be in opts['profiles']

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import logging
import os
import sys

# Import salt libs
import salt.cloud
import salt.config
import salt.defaults.exitcodes
import salt.output
import salt.syspaths as syspaths
import salt.utils.cloud
import salt.utils.parsers
import salt.utils.user
from salt.exceptions import SaltCloudException, SaltCloudSystemExit

# Import 3rd-party libs
from salt.ext import six
from salt.ext.six.moves import input
from salt.utils.verify import check_user, verify_env, verify_log, verify_log_files

log = logging.getLogger(__name__)

# pylint: disable=broad-except


class SaltCloud(salt.utils.parsers.SaltCloudParser):
    def run(self):
        """
        Execute the salt-cloud command line
        """
        # Parse shell arguments
        self.parse_args()

        salt_master_user = self.config.get("user")
        if salt_master_user is None:
            salt_master_user = salt.utils.user.get_user()

        if not check_user(salt_master_user):
            self.error(
                "If salt-cloud is running on a master machine, salt-cloud "
                "needs to run as the same user as the salt-master, '{0}'. "
                "If salt-cloud is not running on a salt-master, the "
                "appropriate write permissions must be granted to '{1}'. "
                "Please run salt-cloud as root, '{0}', or change "
                "permissions for '{1}'.".format(salt_master_user, syspaths.CONFIG_DIR)
            )

        try:
            if self.config["verify_env"]:
                verify_env(
                    [os.path.dirname(self.config["conf_file"])],
                    salt_master_user,
                    root_dir=self.config["root_dir"],
                )
                logfile = self.config["log_file"]
                if logfile is not None:
                    # Logfile is not using Syslog, verify
                    verify_log_files([logfile], salt_master_user)
        except (IOError, OSError) as err:
            log.error("Error while verifying the environment: %s", err)
            sys.exit(err.errno)

        # Setup log file logging
        self.setup_logfile_logger()
        verify_log(self.config)

        if self.options.update_bootstrap:
            ret = salt.utils.cloud.update_bootstrap(self.config)
            salt.output.display_output(ret, self.options.output, opts=self.config)
            self.exit(salt.defaults.exitcodes.EX_OK)

        log.info("salt-cloud starting")
        try:
            mapper = salt.cloud.Map(self.config)
        except SaltCloudSystemExit as exc:
            self.handle_exception(exc.args, exc)
        except SaltCloudException as exc:
            msg = "There was an error generating the mapper."
            self.handle_exception(msg, exc)

        names = self.config.get("names", None)
        if names is not None:
            filtered_rendered_map = {}
            for map_profile in mapper.rendered_map:
                filtered_map_profile = {}
                for name in mapper.rendered_map[map_profile]:
                    if name in names:
                        filtered_map_profile[name] = mapper.rendered_map[map_profile][
                            name
                        ]
                if filtered_map_profile:
                    filtered_rendered_map[map_profile] = filtered_map_profile
            mapper.rendered_map = filtered_rendered_map

        ret = {}

        if self.selected_query_option is not None:
            if self.selected_query_option == "list_providers":
                # pylint: disable=broad-except
                try:
                    ret = mapper.provider_list()
                except (SaltCloudException, Exception,) as exc:
                    msg = "There was an error listing providers: {0}"
                    self.handle_exception(msg, exc)
                # pylint: enable=broad-except

            elif self.selected_query_option == "list_profiles":
                provider = self.options.list_profiles
                # pylint: disable=broad-except
                try:
                    ret = mapper.profile_list(provider)
                except (SaltCloudException, Exception,) as exc:
                    msg = "There was an error listing profiles: {0}"
                    self.handle_exception(msg, exc)
                # pylint: enable=broad-except

            elif self.config.get("map", None):
                log.info("Applying map from '%s'.", self.config["map"])
                # pylint: disable=broad-except
                try:
                    ret = mapper.interpolated_map(query=self.selected_query_option)
                except (SaltCloudException, Exception,) as exc:
                    msg = "There was an error with a custom map: {0}"
                    self.handle_exception(msg, exc)
                # pylint: enable=broad-except
            else:
                # pylint: disable=broad-except
                try:
                    ret = mapper.map_providers_parallel(
                        query=self.selected_query_option
                    )
                except (SaltCloudException, Exception,) as exc:
                    msg = "There was an error with a map: {0}"
                    self.handle_exception(msg, exc)
                # pylint: enable=broad-except

        elif self.options.list_locations is not None:
            # pylint: disable=broad-except
            try:
                ret = mapper.location_list(self.options.list_locations)
            except (SaltCloudException, Exception,) as exc:
                msg = "There was an error listing locations: {0}"
                self.handle_exception(msg, exc)
            # pylint: enable=broad-except

        elif self.options.list_images is not None:
            # pylint: disable=broad-except
            try:
                ret = mapper.image_list(self.options.list_images)
            except (SaltCloudException, Exception,) as exc:
                msg = "There was an error listing images: {0}"
                self.handle_exception(msg, exc)
            # pylint: enable=broad-except

        elif self.options.list_sizes is not None:
            # pylint: disable=broad-except
            try:
                ret = mapper.size_list(self.options.list_sizes)
            except (SaltCloudException, Exception,) as exc:
                msg = "There was an error listing sizes: {0}"
                self.handle_exception(msg, exc)
            # pylint: enable=broad-except

        elif self.options.destroy and (
            self.config.get("names", None) or self.config.get("map", None)
        ):
            map_file = self.config.get("map", None)
            names = self.config.get("names", ())
            if map_file is not None:
                if names != ():
                    msg = (
                        "Supplying a mapfile, '{0}', in addition to instance names {1} "
                        "with the '--destroy' or '-d' function is not supported. "
                        "Please choose to delete either the entire map file or individual "
                        "instances.".format(map_file, names)
                    )
                    self.handle_exception(msg, SaltCloudSystemExit)

                log.info("Applying map from '%s'.", map_file)
                matching = mapper.delete_map(query="list_nodes")
            else:
                matching = mapper.get_running_by_names(
                    names, profile=self.options.profile
                )

            if not matching:
                print("No machines were found to be destroyed")
                self.exit(salt.defaults.exitcodes.EX_OK)

            msg = "The following virtual machines are set to be destroyed:\n"
            names = set()
            for alias, drivers in six.iteritems(matching):
                msg += "  {0}:\n".format(alias)
                for driver, vms in six.iteritems(drivers):
                    msg += "    {0}:\n".format(driver)
                    for name in vms:
                        msg += "      {0}\n".format(name)
                        names.add(name)
            # pylint: disable=broad-except
            try:
                if self.print_confirm(msg):
                    ret = mapper.destroy(names, cached=True)
            except (SaltCloudException, Exception,) as exc:
                msg = "There was an error destroying machines: {0}"
                self.handle_exception(msg, exc)
            # pylint: enable=broad-except

        elif self.options.action and (
            self.config.get("names", None) or self.config.get("map", None)
        ):
            if self.config.get("map", None):
                log.info("Applying map from '%s'.", self.config["map"])
                try:
                    names = mapper.get_vmnames_by_action(self.options.action)
                except SaltCloudException as exc:
                    msg = "There was an error actioning virtual machines."
                    self.handle_exception(msg, exc)
            else:
                names = self.config.get("names", None)

            kwargs = {}
            machines = []
            msg = (
                "The following virtual machines are set to be actioned with "
                '"{0}":\n'.format(self.options.action)
            )
            for name in names:
                if "=" in name:
                    # This is obviously not a machine name, treat it as a kwarg
                    key, value = name.split("=", 1)
                    kwargs[key] = value
                else:
                    msg += "  {0}\n".format(name)
                    machines.append(name)
            names = machines

            # pylint: disable=broad-except
            try:
                if self.print_confirm(msg):
                    ret = mapper.do_action(names, kwargs)
            except (SaltCloudException, Exception,) as exc:
                msg = "There was an error actioning machines: {0}"
                self.handle_exception(msg, exc)
            # pylint: enable=broad-except

        elif self.options.function:
            kwargs = {}
            args = self.args[:]
            for arg in args[:]:
                if "=" in arg:
                    key, value = arg.split("=", 1)
                    kwargs[key] = value
                    args.remove(arg)

            if args:
                self.error(
                    "Any arguments passed to --function need to be passed "
                    "as kwargs. Ex: image=ami-54cf5c3d. Remaining "
                    "arguments: {0}".format(args)
                )
            # pylint: disable=broad-except
            try:
                ret = mapper.do_function(
                    self.function_provider, self.function_name, kwargs
                )
            except (SaltCloudException, Exception,) as exc:
                msg = "There was an error running the function: {0}"
                self.handle_exception(msg, exc)
            # pylint: enable=broad-except

        elif self.options.profile and self.config.get("names", False):
            # pylint: disable=broad-except
            try:
                ret = mapper.run_profile(self.options.profile, self.config.get("names"))
            except (SaltCloudException, Exception,) as exc:
                msg = "There was a profile error: {0}"
                self.handle_exception(msg, exc)
            # pylint: enable=broad-except

        elif self.options.set_password:
            username = self.credential_username
            provider_name = "salt.cloud.provider.{0}".format(self.credential_provider)
            # TODO: check if provider is configured
            # set the password
            salt.utils.cloud.store_password_in_keyring(provider_name, username)
        elif self.config.get("map", None) and self.selected_query_option is None:
            if not mapper.rendered_map:
                sys.stderr.write("No nodes defined in this map")
                self.exit(salt.defaults.exitcodes.EX_GENERIC)
            # pylint: disable=broad-except
            try:
                ret = {}
                run_map = True

                log.info("Applying map from '%s'.", self.config["map"])
                dmap = mapper.map_data()

                msg = ""
                if "errors" in dmap:
                    # display profile errors
                    msg += "Found the following errors:\n"
                    for profile_name, error in six.iteritems(dmap["errors"]):
                        msg += "  {0}: {1}\n".format(profile_name, error)
                    sys.stderr.write(msg)
                    sys.stderr.flush()

                msg = ""
                if "existing" in dmap:
                    msg += "The following virtual machines already exist:\n"
                    for name in dmap["existing"]:
                        msg += "  {0}\n".format(name)

                if dmap["create"]:
                    msg += "The following virtual machines are set to be " "created:\n"
                    for name in dmap["create"]:
                        msg += "  {0}\n".format(name)

                if "destroy" in dmap:
                    msg += (
                        "The following virtual machines are set to be " "destroyed:\n"
                    )
                    for name in dmap["destroy"]:
                        msg += "  {0}\n".format(name)

                if not dmap["create"] and not dmap.get("destroy", None):
                    if not dmap.get("existing", None):
                        # nothing to create or destroy & nothing exists
                        print(msg)
                        self.exit(1)
                    else:
                        # nothing to create or destroy, print existing
                        run_map = False

                if run_map:
                    if self.print_confirm(msg):
                        ret = mapper.run_map(dmap)

                    if self.config.get("parallel", False) is False:
                        log.info("Complete")

                if dmap.get("existing", None):
                    for name in dmap["existing"]:
                        if "ec2" in dmap["existing"][name]["provider"]:
                            msg = "Instance already exists, or is terminated and has the same name."
                        else:
                            msg = "Already running."
                        ret[name] = {"Message": msg}

            except (SaltCloudException, Exception,) as exc:
                msg = "There was a query error: {0}"
                self.handle_exception(msg, exc)
            # pylint: enable=broad-except

        elif self.options.bootstrap:
            host = self.options.bootstrap
            if self.args and "=" not in self.args[0]:
                minion_id = self.args.pop(0)
            else:
                minion_id = host

            vm_ = {
                "driver": "",
                "ssh_host": host,
                "name": minion_id,
            }
            args = self.args[:]
            for arg in args[:]:
                if "=" in arg:
                    key, value = arg.split("=", 1)
                    vm_[key] = value
                    args.remove(arg)

            if args:
                self.error(
                    "Any arguments passed to --bootstrap need to be passed as "
                    "kwargs. Ex: ssh_username=larry. Remaining arguments: {0}".format(
                        args
                    )
                )

            # pylint: disable=broad-except
            try:
                ret = salt.utils.cloud.bootstrap(vm_, self.config)
            except (SaltCloudException, Exception,) as exc:
                msg = "There was an error bootstrapping the minion: {0}"
                self.handle_exception(msg, exc)
            # pylint: enable=broad-except

        else:
            self.error("Nothing was done. Using the proper arguments?")

        salt.output.display_output(ret, self.options.output, opts=self.config)
        self.exit(salt.defaults.exitcodes.EX_OK)

    def print_confirm(self, msg):
        if self.options.assume_yes:
            return True
        print(msg)
        res = input("Proceed? [N/y] ")
        if not res.lower().startswith("y"):
            return False
        print("... proceeding")
        return True

    def handle_exception(self, msg, exc):
        if isinstance(exc, SaltCloudException):
            # It's a known exception and we know how to handle it
            if isinstance(exc, SaltCloudSystemExit):
                # This is a salt cloud system exit
                if exc.exit_code > 0:
                    # the exit code is bigger than 0, it's an error
                    msg = "Error: {0}".format(msg)
                self.exit(exc.exit_code, msg.format(exc).rstrip() + "\n")
            # It's not a system exit but it's an error we can
            # handle
            self.error(msg.format(exc))
        # This is a generic exception, log it, include traceback if
        # debug logging is enabled and exit.
        # pylint: disable=str-format-in-logging
        log.error(
            msg.format(exc),
            # Show the traceback if the debug logging level is
            # enabled
            exc_info_on_loglevel=logging.DEBUG,
        )
        # pylint: enable=str-format-in-logging
        self.exit(salt.defaults.exitcodes.EX_GENERIC)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)


    tests.unit.version_test
    ~~~~~~~~~~~~~~~~~~~~~~~

    Test salt's regex git describe version parsing
"""

# Import python libs
from __future__ import absolute_import

import re

import salt.version

# Import Salt libs
from salt.version import SaltStackVersion, versions_report
from tests.support.mock import MagicMock, patch

# Import Salt Testing libs
from tests.support.unit import TestCase


class VersionTestCase(TestCase):
    def test_version_parsing(self):
        strip_initial_non_numbers_regex = re.compile(r"(?:[^\d]+)?(?P<vs>.*)")
        expect = (
            ("v0.12.0-19-g767d4f9", (0, 12, 0, 0, "", 0, 19, "g767d4f9"), None),
            ("v0.12.0-85-g2880105", (0, 12, 0, 0, "", 0, 85, "g2880105"), None),
            (
                "debian/0.11.1+ds-1-3-ga0afcbd",
                (0, 11, 1, 0, "", 0, 3, "ga0afcbd"),
                "0.11.1-3-ga0afcbd",
            ),
            ("0.12.1", (0, 12, 1, 0, "", 0, 0, None), None),
            ("0.12.1", (0, 12, 1, 0, "", 0, 0, None), None),
            ("0.17.0rc1", (0, 17, 0, 0, "rc", 1, 0, None), None),
            ("v0.17.0rc1-1-g52ebdfd", (0, 17, 0, 0, "rc", 1, 1, "g52ebdfd"), None),
            ("v2014.1.4.1", (2014, 1, 4, 1, "", 0, 0, None), None),
            (
                "v2014.1.4.1rc3-n/a-abcdefff",
                (2014, 1, 4, 1, "rc", 3, -1, "abcdefff"),
                None,
            ),
            ("v3.4.1.1", (3, 4, 1, 1, "", 0, 0, None), None),
            ("v3000", (3000, "", 0, 0, None), "3000"),
            ("v3000.0", (3000, "", 0, 0, None), "3000"),
            ("v4518.1", (4518, 1, "", 0, 0, None), "4518.1"),
            ("v3000rc1", (3000, "rc", 1, 0, None), "3000rc1"),
            ("v3000rc1-n/a-abcdefff", (3000, "rc", 1, -1, "abcdefff"), None),
            ("3000-n/a-1e7bc8f", (3000, "", 0, -1, "1e7bc8f"), None),
            ("3000.1-n/a-1e7bc8f", (3000, 1, "", 0, -1, "1e7bc8f"), None),
        )

        for vstr, full_info, version in expect:
            saltstack_version = SaltStackVersion.parse(vstr)
            self.assertEqual(saltstack_version.full_info, full_info)
            if version is None:
                version = strip_initial_non_numbers_regex.search(vstr).group("vs")
            self.assertEqual(saltstack_version.string, version)

    def test_version_comparison(self):
        examples = (
            ("debian/0.11.1+ds-1-3-ga0afcbd", "0.11.1+ds-2"),
            ("v0.12.0-85-g2880105", "v0.12.0-19-g767d4f9"),
            ("v0.17.0rc1-1-g52ebdfd", "0.17.0rc1"),
            ("v0.17.0", "v0.17.0rc1"),
            ("Hydrogen", "0.17.0"),
            ("Helium", "Hydrogen"),
            ("v2014.1.4.1-n/a-abcdefff", "v2014.1.4.1rc3-n/a-abcdefff"),
            ("v2014.1.4.1-1-abcdefff", "v2014.1.4.1-n/a-abcdefff"),
            ("v2016.12.0rc1", "v2016.12.0b1"),
            ("v2016.12.0beta1", "v2016.12.0alpha1"),
            ("v2016.12.0alpha1", "v2016.12.0alpha0"),
            ("v3000.1", "v3000"),
            ("v3000rc2", "v3000rc1"),
            ("v3001", "v3000"),
            ("v4023rc1", "v4022rc1"),
            ("v3000", "v3000rc1"),
            ("v3000", "v2019.2.1"),
            ("v3000.1", "v2019.2.1"),
            # we created v3000.0rc1 tag on repo
            # but we should not be using this
            # version scheme in the future
            # but still adding test for it
            ("v3000", "v3000.0rc1"),
            ("v3000.1rc1", "v3000.0rc1"),
            ("v3000", "v2019.2.1rc1"),
            ("v3001rc1", "v2019.2.1rc1"),
        )
        for higher_version, lower_version in examples:
            self.assertTrue(SaltStackVersion.parse(higher_version) > lower_version)
            self.assertTrue(SaltStackVersion.parse(lower_version) < higher_version)
            assert SaltStackVersion.parse(lower_version) != higher_version

    def test_unparsable_version(self):
        with self.assertRaises(ValueError):
            SaltStackVersion.from_name("Drunk")

        with self.assertRaises(ValueError):
            SaltStackVersion.parse("Drunk")

    def test_sha(self):
        """
        test matching sha's
        """
        exp_ret = (
            ("d6cd1e2bd19e03a81132a23b2025920577f84e37", True),
            ("2880105", True),
            ("v3000.0.1", False),
            ("v0.12.0-85-g2880105", False),
        )
        for commit, exp in exp_ret:
            ret = SaltStackVersion.git_sha_regex.match(commit)
            if exp:
                assert ret
            else:
                assert not ret

    def test_version_report_lines(self):
        """
        Validate padding in versions report is correct
        """
        # Get a set of all version report name lenghts including padding
        line_lengths = set(
            [
                len(line.split(":")[0])
                for line in list(versions_report())[4:]
                if line != " " and line != "System Versions:"
            ]
        )
        # Check that they are all the same size (only one element in the set)
        assert len(line_lengths) == 1

    def test_string_new_version(self):
        """
        Validate string property method
        using new versioning scheme
        """
        maj_ver = "3000"
        ver = SaltStackVersion(major=maj_ver)
        assert not ver.minor
        assert not ver.bugfix
        assert maj_ver == ver.string

    def test_string_new_version_minor(self):
        """
        Validate string property method
        using new versioning scheme alongside
        minor version
        """
        maj_ver = 3000
        min_ver = 1
        ver = SaltStackVersion(major=maj_ver, minor=min_ver)
        assert ver.minor == min_ver
        assert not ver.bugfix
        assert ver.string == "{0}.{1}".format(maj_ver, min_ver)

    def test_string_new_version_minor_as_string(self):
        """
        Validate string property method
        using new versioning scheme alongside
        minor version
        """
        maj_ver = "3000"
        min_ver = "1"
        ver = SaltStackVersion(major=maj_ver, minor=min_ver)
        assert ver.minor == int(min_ver)
        assert not ver.bugfix
        assert ver.string == "{0}.{1}".format(maj_ver, min_ver)

        # This only seems to happen on a cloned repo without its tags
        maj_ver = "3000"
        min_ver = ""
        ver = SaltStackVersion(major=maj_ver, minor=min_ver)
        assert ver.minor is None, "{!r} is not {!r}".format(ver.minor, min_ver)
        assert not ver.bugfix
        assert ver.string == maj_ver

    def test_string_old_version(self):
        """
        Validate string property method
        using old versioning scheme alongside
        minor version
        """
        maj_ver = "2019"
        min_ver = "2"
        ver = SaltStackVersion(major=maj_ver, minor=min_ver)
        assert ver.bugfix == 0
        assert ver.string == "{0}.{1}.0".format(maj_ver, min_ver)

    def test_noc_info(self):
        """
        Test noc_info property method
        """
        expect = (
            ("v2014.1.4.1rc3-n/a-abcdefff", (2014, 1, 4, 1, "rc", 3, -1)),
            ("v3.4.1.1", (3, 4, 1, 1, "", 0, 0)),
            ("v3000", (3000, "", 0, 0)),
            ("v3000.0", (3000, "", 0, 0)),
            ("v4518.1", (4518, 1, "", 0, 0)),
            ("v3000rc1", (3000, "rc", 1, 0)),
            ("v3000rc1-n/a-abcdefff", (3000, "rc", 1, -1)),
        )

        for vstr, noc_info in expect:
            saltstack_version = SaltStackVersion.parse(vstr)
            assert saltstack_version.noc_info, noc_info
            assert len(saltstack_version.noc_info) == len(noc_info)

    def test_full_info(self):
        """
        Test full_Info property method
        """
        expect = (
            ("v2014.1.4.1rc3-n/a-abcdefff", (2014, 1, 4, 1, "rc", 3, -1, "abcdefff")),
            ("v3.4.1.1", (3, 4, 1, 1, "", 0, 0, None)),
            ("v3000", (3000, "", 0, 0, None)),
            ("v3000.0", (3000, "", 0, 0, None)),
            ("v4518.1", (4518, 1, "", 0, 0, None)),
            ("v3000rc1", (3000, "rc", 1, 0, None)),
            ("v3000rc1-n/a-abcdefff", (3000, "rc", 1, -1, "abcdefff")),
        )

        for vstr, full_info in expect:
            saltstack_version = SaltStackVersion.parse(vstr)
            assert saltstack_version.full_info, full_info
            assert len(saltstack_version.full_info) == len(full_info)

    def test_full_info_all_versions(self):
        """
        Test full_info_all_versions property method
        """
        expect = (
            ("v2014.1.4.1rc3-n/a-abcdefff", (2014, 1, 4, 1, "rc", 3, -1, "abcdefff")),
            ("v3.4.1.1", (3, 4, 1, 1, "", 0, 0, None)),
            ("v3000", (3000, None, None, 0, "", 0, 0, None)),
            ("v3000.0", (3000, 0, None, 0, "", 0, 0, None)),
            ("v4518.1", (4518, 1, None, 0, "", 0, 0, None)),
            ("v3000rc1", (3000, None, None, 0, "rc", 2, 0, None)),
            ("v3000rc1-n/a-abcdefff", (3000, None, None, 0, "rc", 1, -1, "abcdefff")),
        )

        for vstr, full_info in expect:
            saltstack_version = SaltStackVersion.parse(vstr)
            assert saltstack_version.full_info_all_versions, full_info
            assert len(saltstack_version.full_info_all_versions) == len(full_info)

    def test_discover_version(self):
        """
        Test call to __discover_version
        when using different versions
        """
        version = {
            ("3000", None): {
                (b"v3000.0rc2-12-g44fe283a77\n", "3000rc2-12-g44fe283a77"),
                (b"v3000", "3000"),
                (b"1234567", "3000-n/a-1234567"),
            },
            (2019, 2): {
                (b"v2019.2.0rc2-12-g44fe283a77\n", "2019.2.0rc2-12-g44fe283a77"),
                (b"v2019.2.0", "2019.2.0"),
                (b"afc9830198dj", "2019.2.0-n/a-afc9830198dj"),
            },
        }
        for maj_min, test_v in version.items():
            for tag_ver, exp in version[maj_min]:
                salt_ver = SaltStackVersion(
                    major=maj_min[0], minor=maj_min[1], bugfix=None
                )
                attrs = {
                    "communicate.return_value": (tag_ver, b""),
                    "returncode.return_value": 0,
                }
                proc_ret = MagicMock(**attrs)
                proc_mock = patch("subprocess.Popen", return_value=proc_ret)
                patch_os = patch("os.path.exists", return_value=True)

                with proc_mock, patch_os:
                    ret = getattr(salt.version, "__discover_version")(salt_ver)
                assert ret == exp

    def test_info_new_version(self):
        """
        test info property method with new versioning scheme
        """
        vers = ((3000, None, None), (3000, 1, None), (3001, 0, None))
        for maj_ver, min_ver, bug_fix in vers:
            ver = SaltStackVersion(major=maj_ver, minor=min_ver, bugfix=bug_fix)
            if min_ver:
                assert ver.info == (maj_ver, min_ver)
            else:
                assert ver.info == (maj_ver,)

    def test_info_old_version(self):
        """
        test info property method with old versioning scheme
        """
        vers = ((2019, 2, 1), (2018, 3, 0), (2017, 7, None))
        for maj_ver, min_ver, bug_fix in vers:
            ver = SaltStackVersion(major=maj_ver, minor=min_ver, bugfix=bug_fix)
            if bug_fix is None:
                assert ver.info == (maj_ver, min_ver, 0, 0)
            else:
                assert ver.info == (maj_ver, min_ver, bug_fix, 0)

    def test_bugfix_string(self):
        """
        test when bugfix is an empty string
        """
        ret = SaltStackVersion(3000, 1, "", 0, 0, None)
        assert ret.info == (3000, 1)
        assert ret.minor == 1
        assert ret.bugfix is None

    def test_version_repr(self):
        """
        Test SaltStackVersion repr for both date
        and new versioning scheme
        """
        expect = (
            (
                (3000, 1, None, None, "", 0, 0, None),
                "<SaltStackVersion name='Neon' major=3000 minor=1>",
            ),
            (
                (3000, 0, None, None, "", 0, 0, None),
                "<SaltStackVersion name='Neon' major=3000>",
            ),
            (
                (2019, 2, 3, None, "", 0, 0, None),
                "<SaltStackVersion name='Fluorine' major=2019 minor=2 bugfix=3>",
            ),
            (
                (2019, 2, 3, None, "rc", 1, 0, None),
                "<SaltStackVersion name='Fluorine' major=2019 minor=2 bugfix=3 rc=1>",
            ),
        )

        for ver, repr_ret in expect:
            assert repr(SaltStackVersion(*ver)) == repr_ret



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.smtp as smtp

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class SmtpTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.states.smtp
    """

    def setup_loader_modules(self):
        return {smtp: {}}

    # 'send_msg' function tests: 1

    def test_send_msg(self):
        """
        Test to send a message via SMTP
        """
        name = "This is a salt states module"

        comt = (
            "Need to send message to admin@example.com:" " This is a salt states module"
        )

        ret = {"name": name, "changes": {}, "result": None, "comment": comt}

        with patch.dict(smtp.__opts__, {"test": True}):
            self.assertDictEqual(
                smtp.send_msg(
                    name,
                    "admin@example.com",
                    "Message from Salt",
                    "admin@example.com",
                    "my-smtp-account",
                ),
                ret,
            )

        comt = "Sent message to admin@example.com: " "This is a salt states module"

        with patch.dict(smtp.__opts__, {"test": False}):
            mock = MagicMock(return_value=True)
            with patch.dict(smtp.__salt__, {"smtp.send_msg": mock}):
                ret["comment"] = comt
                ret["result"] = True
                self.assertDictEqual(
                    smtp.send_msg(
                        name,
                        "admin@example.com",
                        "Message from Salt",
                        "admin@example.com",
                        "my-smtp-account",
                    ),
                    ret,
                )



# -*- coding: utf-8 -*-
"""
Simple Smoke Tests for Connected Proxy Minion
"""
from __future__ import absolute_import, print_function, unicode_literals

from tests.support.case import ModuleCase
from tests.support.helpers import slowTest


class ProxyMinionSimpleTestCase(ModuleCase):
    """
    Test proxy minion functionality
    """

    def test_can_it_ping(self):
        """
        Ensure the proxy can ping
        """
        ret = self.run_function("test.ping", minion_tgt="proxytest")
        self.assertEqual(ret, True)

    def test_list_pkgs(self):
        """
        Package test 1, really just tests that the virtual function capability
        is working OK.
        """
        ret = self.run_function("pkg.list_pkgs", minion_tgt="proxytest")
        self.assertIn("coreutils", ret)
        self.assertIn("apache", ret)
        self.assertIn("redbull", ret)

    def test_install_pkgs(self):
        """
        Package test 2, really just tests that the virtual function capability
        is working OK.
        """
        ret = self.run_function("pkg.install", ["thispkg"], minion_tgt="proxytest")
        self.assertEqual(ret["thispkg"], "1.0")

        ret = self.run_function("pkg.list_pkgs", minion_tgt="proxytest")

        self.assertEqual(ret["apache"], "2.4")
        self.assertEqual(ret["redbull"], "999.99")
        self.assertEqual(ret["thispkg"], "1.0")

    def test_remove_pkgs(self):
        ret = self.run_function("pkg.remove", ["apache"], minion_tgt="proxytest")
        self.assertNotIn("apache", ret)

    def test_upgrade(self):
        ret = self.run_function("pkg.upgrade", minion_tgt="proxytest")
        self.assertEqual(ret["coreutils"]["new"], "2.0")
        self.assertEqual(ret["redbull"]["new"], "1000.99")

    def test_service_list(self):
        ret = self.run_function("service.list", minion_tgt="proxytest")
        self.assertIn("ntp", ret)

    def test_service_stop(self):
        ret = self.run_function("service.stop", ["ntp"], minion_tgt="proxytest")
        ret = self.run_function("service.status", ["ntp"], minion_tgt="proxytest")
        self.assertFalse(ret)

    def test_service_start(self):
        ret = self.run_function("service.start", ["samba"], minion_tgt="proxytest")
        ret = self.run_function("service.status", ["samba"], minion_tgt="proxytest")
        self.assertTrue(ret)

    def test_service_get_all(self):
        ret = self.run_function("service.get_all", minion_tgt="proxytest")
        self.assertTrue(ret)
        self.assertIn("samba", " ".join(ret))

    def test_grains_items(self):
        ret = self.run_function("grains.items", minion_tgt="proxytest")
        self.assertEqual(ret["kernel"], "proxy")
        self.assertEqual(ret["kernelrelease"], "proxy")

    def test_state_apply(self):
        ret = self.run_function("state.apply", ["core"], minion_tgt="proxytest")
        for key, value in ret.items():
            self.assertTrue(value["result"])

    @slowTest
    def test_state_highstate(self):
        ret = self.run_function("state.highstate", minion_tgt="proxytest")
        for key, value in ret.items():
            self.assertTrue(value["result"])



# -*- coding: utf-8 -*-
"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""
# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

import os

# Import 3rd-party libs
import jinja2.exceptions
import salt.modules.rh_ip as rh_ip
from salt.ext import six

# Import Salt Libs
from salt.ext.six.moves import range

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class RhipTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.rh_ip
    """

    def setup_loader_modules(self):
        return {rh_ip: {}}

    def test_error_message_iface_should_process_non_str_expected(self):
        values = [1, True, False, "no-kaboom"]
        iface = "ethtest"
        option = "test"
        msg = rh_ip._error_msg_iface(iface, option, values)
        self.assertTrue(msg.endswith("[1|True|False|no-kaboom]"), msg)

    def test_error_message_network_should_process_non_str_expected(self):
        values = [1, True, False, "no-kaboom"]
        msg = rh_ip._error_msg_network("fnord", values)
        self.assertTrue(msg.endswith("[1|True|False|no-kaboom]"), msg)

    def test_build_bond(self):
        """
        Test to create a bond script in /etc/modprobe.d with the passed
        settings and load the bonding kernel module.
        """
        with patch.dict(rh_ip.__grains__, {"osrelease": "osrelease"}):
            with patch.object(rh_ip, "_parse_settings_bond", MagicMock()):
                mock = jinja2.exceptions.TemplateNotFound("foo")
                with patch.object(
                    jinja2.Environment, "get_template", MagicMock(side_effect=mock)
                ):
                    self.assertEqual(rh_ip.build_bond("iface"), "")

                with patch.dict(
                    rh_ip.__salt__, {"kmod.load": MagicMock(return_value=None)}
                ):
                    with patch.object(rh_ip, "_write_file_iface", return_value=None):
                        with patch.object(rh_ip, "_read_temp", return_value="A"):
                            self.assertEqual(rh_ip.build_bond("iface", test="A"), "A")

                        with patch.object(rh_ip, "_read_file", return_value="A"):
                            self.assertEqual(rh_ip.build_bond("iface", test=None), "A")

    def test_build_interface(self):
        """
        Test to build an interface script for a network interface.
        """
        with patch.dict(rh_ip.__grains__, {"os": "Fedora", "osmajorrelease": 26}):
            with patch.object(rh_ip, "_raise_error_iface", return_value=None):
                self.assertRaises(
                    AttributeError, rh_ip.build_interface, "iface", "slave", True
                )

                with patch.dict(
                    rh_ip.__salt__, {"network.interfaces": lambda: {"eth": True}}
                ):
                    self.assertRaises(
                        AttributeError,
                        rh_ip.build_interface,
                        "iface",
                        "eth",
                        True,
                        netmask="255.255.255.255",
                        prefix=32,
                        test=True,
                    )
                    self.assertRaises(
                        AttributeError,
                        rh_ip.build_interface,
                        "iface",
                        "eth",
                        True,
                        ipaddrs=["A"],
                        test=True,
                    )
                    self.assertRaises(
                        AttributeError,
                        rh_ip.build_interface,
                        "iface",
                        "eth",
                        True,
                        ipv6addrs=["A"],
                        test=True,
                    )

        for osrelease in range(5, 8):
            with patch.dict(
                rh_ip.__grains__,
                {"os": "RedHat", "osrelease": six.text_type(osrelease)},
            ):
                with patch.object(rh_ip, "_raise_error_iface", return_value=None):
                    with patch.object(rh_ip, "_parse_settings_bond", MagicMock()):
                        mock = jinja2.exceptions.TemplateNotFound("foo")
                        with patch.object(
                            jinja2.Environment,
                            "get_template",
                            MagicMock(side_effect=mock),
                        ):
                            self.assertEqual(
                                rh_ip.build_interface("iface", "vlan", True), ""
                            )

                        with patch.object(rh_ip, "_read_temp", return_value="A"):
                            with patch.object(
                                jinja2.Environment, "get_template", MagicMock()
                            ):
                                self.assertEqual(
                                    rh_ip.build_interface(
                                        "iface", "vlan", True, test="A"
                                    ),
                                    "A",
                                )

                                with patch.object(
                                    rh_ip, "_write_file_iface", return_value=None
                                ):
                                    with patch.object(
                                        os.path, "join", return_value="A"
                                    ):
                                        with patch.object(
                                            rh_ip, "_read_file", return_value="A"
                                        ):
                                            self.assertEqual(
                                                rh_ip.build_interface(
                                                    "iface", "vlan", True
                                                ),
                                                "A",
                                            )
                                            if osrelease > 6:
                                                with patch.dict(
                                                    rh_ip.__salt__,
                                                    {
                                                        "network.interfaces": lambda: {
                                                            "eth": True
                                                        }
                                                    },
                                                ):
                                                    self.assertEqual(
                                                        rh_ip.build_interface(
                                                            "iface",
                                                            "eth",
                                                            True,
                                                            ipaddrs=["127.0.0.1/8"],
                                                        ),
                                                        "A",
                                                    )
                                                    self.assertEqual(
                                                        rh_ip.build_interface(
                                                            "iface",
                                                            "eth",
                                                            True,
                                                            ipv6addrs=["fc00::1/128"],
                                                        ),
                                                        "A",
                                                    )

    def test_build_routes(self):
        """
        Test to build a route script for a network interface.
        """
        with patch.dict(rh_ip.__grains__, {"osrelease": "5.0"}):
            with patch.object(rh_ip, "_parse_routes", MagicMock()):
                mock = jinja2.exceptions.TemplateNotFound("foo")
                with patch.object(
                    jinja2.Environment, "get_template", MagicMock(side_effect=mock)
                ):
                    self.assertEqual(rh_ip.build_routes("iface"), "")

                with patch.object(jinja2.Environment, "get_template", MagicMock()):
                    with patch.object(rh_ip, "_read_temp", return_value=["A"]):
                        self.assertEqual(rh_ip.build_routes("i", test="t"), ["A", "A"])

                    with patch.object(rh_ip, "_read_file", return_value=["A"]):
                        with patch.object(os.path, "join", return_value="A"):
                            with patch.object(
                                rh_ip, "_write_file_iface", return_value=None
                            ):
                                self.assertEqual(
                                    rh_ip.build_routes("i", test=None), ["A", "A"]
                                )

    def test_down(self):
        """
        Test to shutdown a network interface
        """
        with patch.dict(rh_ip.__salt__, {"cmd.run": MagicMock(return_value="A")}):
            self.assertEqual(rh_ip.down("iface", "iface_type"), "A")

        self.assertEqual(rh_ip.down("iface", "slave"), None)

    def test_get_bond(self):
        """
        Test to return the content of a bond script
        """
        with patch.object(os.path, "join", return_value="A"):
            with patch.object(rh_ip, "_read_file", return_value="A"):
                self.assertEqual(rh_ip.get_bond("iface"), "A")

    def test_get_interface(self):
        """
        Test to return the contents of an interface script
        """
        with patch.object(os.path, "join", return_value="A"):
            with patch.object(rh_ip, "_read_file", return_value="A"):
                self.assertEqual(rh_ip.get_interface("iface"), "A")

    def test__parse_settings_eth_hwaddr_and_macaddr(self):
        """
        Test that an AttributeError is thrown when hwaddr and macaddr are
        passed together. They cannot be used together
        """
        opts = {"hwaddr": 1, "macaddr": 2}

        self.assertRaises(
            AttributeError,
            rh_ip._parse_settings_eth,
            opts=opts,
            iface_type="eth",
            enabled=True,
            iface="eth0",
        )

    def test__parse_settings_eth_hwaddr(self):
        """
        Make sure hwaddr gets added when parsing opts
        """
        opts = {"hwaddr": "AA:BB:CC:11:22:33"}
        with patch.dict(rh_ip.__salt__, {"network.interfaces": MagicMock()}):
            results = rh_ip._parse_settings_eth(
                opts=opts, iface_type="eth", enabled=True, iface="eth0"
            )
        self.assertIn("hwaddr", results)
        self.assertEqual(results["hwaddr"], opts["hwaddr"])

    def test__parse_settings_eth_macaddr(self):
        """
        Make sure macaddr gets added when parsing opts
        """
        opts = {"macaddr": "AA:BB:CC:11:22:33"}
        with patch.dict(rh_ip.__salt__, {"network.interfaces": MagicMock()}):
            results = rh_ip._parse_settings_eth(
                opts=opts, iface_type="eth", enabled=True, iface="eth0"
            )
        self.assertIn("macaddr", results)
        self.assertEqual(results["macaddr"], opts["macaddr"])

    def test_up(self):
        """
        Test to start up a network interface
        """
        with patch.dict(rh_ip.__salt__, {"cmd.run": MagicMock(return_value="A")}):
            self.assertEqual(rh_ip.up("iface", "iface_type"), "A")

        self.assertEqual(rh_ip.up("iface", "slave"), None)

    def test_get_routes(self):
        """
        Test to return the contents of the interface routes script.
        """
        with patch.object(os.path, "join", return_value="A"):
            with patch.object(rh_ip, "_read_file", return_value=["A"]):
                self.assertEqual(rh_ip.get_routes("iface"), ["A", "A"])

    def test_get_network_settings(self):
        """
        Test to return the contents of the global network script.
        """
        with patch.object(rh_ip, "_read_file", return_value="A"):
            self.assertEqual(rh_ip.get_network_settings(), "A")

    def test_apply_network_settings(self):
        """
        Test to apply global network configuration.
        """
        with patch.dict(
            rh_ip.__salt__, {"service.restart": MagicMock(return_value=True)}
        ):
            self.assertTrue(rh_ip.apply_network_settings())

    def test_build_network_settings(self):
        """
        Test to build the global network script.
        """
        with patch.object(rh_ip, "_parse_rh_config", MagicMock()):
            with patch.object(rh_ip, "_parse_network_settings", MagicMock()):

                mock = jinja2.exceptions.TemplateNotFound("foo")
                with patch.object(
                    jinja2.Environment, "get_template", MagicMock(side_effect=mock)
                ):
                    self.assertEqual(rh_ip.build_network_settings(), "")

                with patch.object(jinja2.Environment, "get_template", MagicMock()):
                    with patch.object(rh_ip, "_read_temp", return_value="A"):
                        self.assertEqual(rh_ip.build_network_settings(test="t"), "A")

                        with patch.object(
                            rh_ip, "_write_file_network", return_value=None
                        ):
                            with patch.object(rh_ip, "_read_file", return_value="A"):
                                self.assertEqual(
                                    rh_ip.build_network_settings(test=None), "A"
                                )



# -*- coding: utf-8 -*-
"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.states.win_servermanager as win_servermanager

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class WinServermanagerTestCase(TestCase, LoaderModuleMockMixin):
    """
        Validate the win_servermanager state
    """

    def setup_loader_modules(self):
        return {win_servermanager: {}}

    def test_installed(self):
        """
            Test to install the windows feature
        """
        mock_list = MagicMock(
            side_effect=[
                {"spongebob": "squarepants"},
                {"squidward": "patrick"},
                {"spongebob": "squarepants"},
                {"spongebob": "squarepants", "squidward": "patrick"},
            ]
        )
        mock_install = MagicMock(
            return_value={
                "Success": True,
                "Restarted": False,
                "RestartNeeded": False,
                "ExitCode": 1234,
                "Features": {
                    "squidward": {
                        "DisplayName": "Squidward",
                        "Message": "",
                        "RestartNeeded": True,
                        "SkipReason": 0,
                        "Success": True,
                    }
                },
            }
        )
        with patch.dict(
            win_servermanager.__salt__,
            {
                "win_servermanager.list_installed": mock_list,
                "win_servermanager.install": mock_install,
            },
        ):
            ret = {
                "name": "spongebob",
                "changes": {},
                "result": True,
                "comment": "The following features are already installed:\n"
                "- spongebob",
            }
            self.assertDictEqual(win_servermanager.installed("spongebob"), ret)

            with patch.dict(win_servermanager.__opts__, {"test": True}):
                ret = {
                    "name": "spongebob",
                    "result": None,
                    "comment": "",
                    "changes": {"spongebob": "Will be installed recurse=False"},
                }
                self.assertDictEqual(win_servermanager.installed("spongebob"), ret)

            with patch.dict(win_servermanager.__opts__, {"test": False}):
                ret = {
                    "name": "squidward",
                    "result": True,
                    "comment": "Installed the following:\n- squidward",
                    "changes": {"squidward": {"new": "patrick", "old": ""}},
                }
                self.assertDictEqual(win_servermanager.installed("squidward"), ret)

    def test_removed(self):
        """
            Test to remove the windows feature
        """
        mock_list = MagicMock(
            side_effect=[
                {"spongebob": "squarepants"},
                {"squidward": "patrick"},
                {"spongebob": "squarepants", "squidward": "patrick"},
                {"spongebob": "squarepants"},
            ]
        )
        mock_remove = MagicMock(
            return_value={
                "Success": True,
                "RestartNeeded": False,
                "Restarted": False,
                "ExitCode": 1234,
                "Features": {
                    "squidward": {
                        "DisplayName": "Squidward",
                        "Message": "",
                        "RestartNeeded": True,
                        "SkipReason": 0,
                        "Success": True,
                    }
                },
            }
        )
        with patch.dict(
            win_servermanager.__salt__,
            {
                "win_servermanager.list_installed": mock_list,
                "win_servermanager.remove": mock_remove,
            },
        ):
            ret = {
                "name": "squidward",
                "changes": {},
                "result": True,
                "comment": "The following features are not installed:\n" "- squidward",
            }
            self.assertDictEqual(win_servermanager.removed("squidward"), ret)

            with patch.dict(win_servermanager.__opts__, {"test": True}):
                ret = {
                    "name": "squidward",
                    "result": None,
                    "comment": "",
                    "changes": {"squidward": "Will be removed"},
                }
                self.assertDictEqual(win_servermanager.removed("squidward"), ret)

            with patch.dict(win_servermanager.__opts__, {"test": False}):
                ret = {
                    "name": "squidward",
                    "result": True,
                    "comment": "Removed the following:\n- squidward",
                    "changes": {"squidward": {"new": "", "old": "patrick"}},
                }
                self.assertDictEqual(win_servermanager.removed("squidward"), ret)



# -*- coding: utf-8 -*-
"""
    :codeauthor: Joao Mesquita <jmesquita@sangoma.com>
"""

# Import Python libs
from __future__ import absolute_import, print_function, unicode_literals

from salt import fileserver

# Import Salt Testing libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.unit import TestCase


class MapDiffTestCase(TestCase):
    def test_diff_with_diffent_keys(self):
        """
        Test that different maps are indeed reported different
        """
        map1 = {"file1": 1234}
        map2 = {"file2": 1234}
        assert fileserver.diff_mtime_map(map1, map2) is True

    def test_diff_with_diffent_values(self):
        """
        Test that different maps are indeed reported different
        """
        map1 = {"file1": 12345}
        map2 = {"file1": 1234}
        assert fileserver.diff_mtime_map(map1, map2) is True


class VCSBackendWhitelistCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {fileserver: {}}

    def test_whitelist(self):
        opts = {
            "fileserver_backend": ["roots", "git", "hgfs", "svn"],
            "extension_modules": "",
        }
        fs = fileserver.Fileserver(opts)
        assert fs.servers.whitelist == [
            "git",
            "gitfs",
            "hg",
            "hgfs",
            "svn",
            "svnfs",
            "roots",
        ], fs.servers.whitelist



# -*- coding: utf-8 -*-
"""
Tests for salt.utils.jid
"""

# Import Python libs
from __future__ import absolute_import, unicode_literals

import datetime
import os

# Import Salt libs
import salt.utils.jid
from tests.support.mock import patch
from tests.support.unit import TestCase


class JidTestCase(TestCase):
    def test_jid_to_time(self):
        test_jid = 20131219110700123489
        expected_jid = "2013, Dec 19 11:07:00.123489"
        self.assertEqual(salt.utils.jid.jid_to_time(test_jid), expected_jid)

        # Test incorrect lengths
        incorrect_jid_length = 2012
        self.assertEqual(salt.utils.jid.jid_to_time(incorrect_jid_length), "")

    def test_is_jid(self):
        self.assertTrue(salt.utils.jid.is_jid("20131219110700123489"))  # Valid JID
        self.assertFalse(salt.utils.jid.is_jid(20131219110700123489))  # int
        self.assertFalse(
            salt.utils.jid.is_jid("2013121911070012348911111")
        )  # Wrong length

    def test_gen_jid(self):
        now = datetime.datetime(2002, 12, 25, 12, 0, 0, 0)
        with patch("salt.utils.jid._utc_now", return_value=now):
            ret = salt.utils.jid.gen_jid({})
            self.assertEqual(ret, "20021225120000000000")
            salt.utils.jid.LAST_JID_DATETIME = None
            ret = salt.utils.jid.gen_jid({"unique_jid": True})
            self.assertEqual(ret, "20021225120000000000_{0}".format(os.getpid()))
            ret = salt.utils.jid.gen_jid({"unique_jid": True})
            self.assertEqual(ret, "20021225120000000001_{0}".format(os.getpid()))

    def test_gen_jid_utc(self):
        utcnow = datetime.datetime(2002, 12, 25, 12, 7, 0, 0)
        with patch("salt.utils.jid._utc_now", return_value=utcnow):
            ret = salt.utils.jid.gen_jid({"utc_jid": True})
            self.assertEqual(ret, "20021225120700000000")

    def test_gen_jid_utc_unique(self):
        utcnow = datetime.datetime(2002, 12, 25, 12, 7, 0, 0)
        with patch("salt.utils.jid._utc_now", return_value=utcnow):
            salt.utils.jid.LAST_JID_DATETIME = None
            ret = salt.utils.jid.gen_jid({"utc_jid": True, "unique_jid": True})
            self.assertEqual(ret, "20021225120700000000_{0}".format(os.getpid()))
            ret = salt.utils.jid.gen_jid({"utc_jid": True, "unique_jid": True})
            self.assertEqual(ret, "20021225120700000001_{0}".format(os.getpid()))



# -*- coding: utf-8 -*-
"""
Unit tests for the docker state
"""

# Import Python Libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt Libs
import salt.modules.dockermod as docker_mod
import salt.states.docker_volume as docker_state

# Import Salt Testing Libs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import Mock, patch
from tests.support.unit import TestCase


class DockerVolumeTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test docker_volume states
    """

    def setup_loader_modules(self):
        return {
            docker_mod: {"__context__": {"docker.docker_version": ""}},
            docker_state: {"__opts__": {"test": False}},
        }

    def test_present(self):
        """
        Test docker_volume.present
        """
        volumes = []
        default_driver = "dummy_default"

        def create_volume(name, driver=None, driver_opts=None):
            for v in volumes:
                # present should never try to add a conflicting
                # volume
                self.assertNotEqual(v["Name"], name)
            if driver is None:
                driver = default_driver
            new = {"Name": name, "Driver": driver}
            volumes.append(new)
            return new

        def remove_volume(name):
            old_len = len(volumes)
            removed = [v for v in volumes if v["Name"] == name]
            # present should not have tried to remove a volume
            # that didn't exist
            self.assertEqual(1, len(removed))
            volumes.remove(removed[0])
            return removed[0]

        docker_create_volume = Mock(side_effect=create_volume)
        __salt__ = {
            "docker.create_volume": docker_create_volume,
            "docker.volumes": Mock(return_value={"Volumes": volumes}),
            "docker.remove_volume": Mock(side_effect=remove_volume),
        }
        with patch.dict(docker_state.__dict__, {"__salt__": __salt__}):
            ret = docker_state.present("volume_foo")
            docker_create_volume.assert_called_with(
                "volume_foo", driver=None, driver_opts=None
            )
            self.assertEqual(
                {
                    "name": "volume_foo",
                    "comment": "",
                    "changes": {
                        "created": {"Driver": default_driver, "Name": "volume_foo"},
                    },
                    "result": True,
                },
                ret,
            )
            self.assertEqual(len(volumes), 1)
            self.assertEqual(volumes[0]["Name"], "volume_foo")
            self.assertIs(volumes[0]["Driver"], default_driver)

            # run it again with the same arguments
            orig_volumes = [volumes[0].copy()]
            ret = docker_state.present("volume_foo")
            self.assertEqual(
                {
                    "name": "volume_foo",
                    "comment": "Volume 'volume_foo' already exists.",
                    "changes": {},
                    "result": True,
                },
                ret,
            )
            self.assertEqual(orig_volumes, volumes)

            # run it again with a different driver but don't force
            ret = docker_state.present("volume_foo", driver="local")
            self.assertEqual(
                {
                    "name": "volume_foo",
                    "comment": (
                        "Driver for existing volume 'volume_foo'"
                        " ('dummy_default') does not match specified"
                        " driver ('local') and force is False"
                    ),
                    "changes": {},
                    "result": False,
                },
                ret,
            )
            self.assertEqual(orig_volumes, volumes)

            # run it again with a different driver and force
            ret = docker_state.present("volume_foo", driver="local", force=True)
            self.assertEqual(
                {
                    "name": "volume_foo",
                    "comment": "",
                    "changes": {
                        "removed": {"Driver": default_driver, "Name": "volume_foo"},
                        "created": {"Driver": "local", "Name": "volume_foo"},
                    },
                    "result": True,
                },
                ret,
            )
            mod_orig_volumes = [orig_volumes[0].copy()]
            mod_orig_volumes[0]["Driver"] = "local"
            self.assertEqual(mod_orig_volumes, volumes)

    def test_present_with_another_driver(self):
        """
        Test docker_volume.present
        """
        docker_create_volume = Mock(return_value="created")
        docker_remove_volume = Mock(return_value="removed")
        __salt__ = {
            "docker.create_volume": docker_create_volume,
            "docker.remove_volume": docker_remove_volume,
            "docker.volumes": Mock(
                return_value={"Volumes": [{"Name": "volume_foo", "Driver": "foo"}]}
            ),
        }
        with patch.dict(docker_state.__dict__, {"__salt__": __salt__}):
            ret = docker_state.present("volume_foo", driver="bar", force=True,)
        docker_remove_volume.assert_called_with("volume_foo")
        docker_create_volume.assert_called_with(
            "volume_foo", driver="bar", driver_opts=None
        )
        self.assertEqual(
            ret,
            {
                "name": "volume_foo",
                "comment": "",
                "changes": {"created": "created", "removed": "removed"},
                "result": True,
            },
        )

    def test_present_wo_existing_volumes(self):
        """
        Test docker_volume.present without existing volumes.
        """
        docker_create_volume = Mock(return_value="created")
        docker_remove_volume = Mock(return_value="removed")
        __salt__ = {
            "docker.create_volume": docker_create_volume,
            "docker.remove_volume": docker_remove_volume,
            "docker.volumes": Mock(return_value={"Volumes": None}),
        }
        with patch.dict(docker_state.__dict__, {"__salt__": __salt__}):
            ret = docker_state.present("volume_foo", driver="bar", force=True,)
        docker_create_volume.assert_called_with(
            "volume_foo", driver="bar", driver_opts=None
        )
        self.assertEqual(
            ret,
            {
                "name": "volume_foo",
                "comment": "",
                "changes": {"created": "created"},
                "result": True,
            },
        )

    def test_absent(self):
        """
        Test docker_volume.absent
        """
        docker_remove_volume = Mock(return_value="removed")
        __salt__ = {
            "docker.remove_volume": docker_remove_volume,
            "docker.volumes": Mock(return_value={"Volumes": [{"Name": "volume_foo"}]}),
        }
        with patch.dict(docker_state.__dict__, {"__salt__": __salt__}):
            ret = docker_state.absent("volume_foo",)
        docker_remove_volume.assert_called_with("volume_foo")
        self.assertEqual(
            ret,
            {
                "name": "volume_foo",
                "comment": "",
                "changes": {"removed": "removed"},
                "result": True,
            },
        )



# -*- coding: utf-8 -*-
"""
Simple Smoke Tests for Connected SSH minions
"""
from __future__ import absolute_import, print_function, unicode_literals

from tests.support.case import SSHCase
from tests.support.helpers import requires_system_grains, skip_if_not_root, slowTest


class SSHMasterTestCase(SSHCase):
    """
    Test ssh master functionality
    """

    @slowTest
    def test_can_it_ping(self):
        """
        Ensure the proxy can ping
        """
        ret = self.run_function("test.ping")
        self.assertEqual(ret, True)

    @requires_system_grains
    @skip_if_not_root
    @slowTest
    def test_service(self, grains):
        service = "cron"
        os_family = grains["os_family"]
        os_release = grains["osrelease"]
        if os_family == "RedHat":
            service = "crond"
        elif os_family == "Arch":
            service = "sshd"
        elif os_family == "MacOS":
            service = "org.ntp.ntpd"
            if int(os_release.split(".")[1]) >= 13:
                service = "com.apple.AirPlayXPCHelper"
        self.run_function("service.enable", [service])
        ret = self.run_function("service.get_all")
        self.assertIn(service, ret)
        self.run_function("service.stop", [service])
        ret = self.run_function("service.status", [service])
        self.assertFalse(ret)
        self.run_function("service.start", [service])
        ret = self.run_function("service.status", [service])
        self.assertTrue(ret)

    @requires_system_grains
    @slowTest
    def test_grains_items(self, grains):
        os_family = grains["os_family"]
        ret = self.run_function("grains.items")
        if os_family == "MacOS":
            self.assertEqual(ret["kernel"], "Darwin")
        else:
            self.assertEqual(ret["kernel"], "Linux")

    @slowTest
    def test_state_apply(self):
        ret = self.run_function("state.apply", ["core"])
        for key, value in ret.items():
            self.assertTrue(value["result"])

    @slowTest
    def test_state_highstate(self):
        ret = self.run_function("state.highstate")
        for key, value in ret.items():
            self.assertTrue(value["result"])



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals

import pytest
from salt.ext import six
from tests.support.case import ModuleCase
from tests.support.helpers import slowTest


@pytest.mark.windows_whitelisted
class SysModuleTest(ModuleCase):
    """
    Validate the sys module
    """

    @slowTest
    def test_valid_docs(self):
        """
        Make sure no functions are exposed that don't have valid docstrings
        """
        ret = self.run_function("runtests_helpers.get_invalid_docs")
        if ret == {"missing_docstring": [], "missing_cli_example": []}:
            return

        if isinstance(ret, six.string_types):
            self.fail(ret)

        self.fail(
            "There are some functions which do not have a docstring or do not "
            "have an example:\nNo docstring:\n{0}\nNo example:\n{1}\n".format(
                "\n".join(["  - {0}".format(f) for f in ret["missing_docstring"]]),
                "\n".join(["  - {0}".format(f) for f in ret["missing_cli_example"]]),
            )
        )



# -*- coding: utf-8 -*-
"""
The service module for macOS

.. versionadded:: 2016.3.0

This module has support for services in the following locations.

.. code-block:: bash

    /System/Library/LaunchDaemons/
    /System/Library/LaunchAgents/
    /Library/LaunchDaemons/
    /Library/LaunchAgents/

    # As of version "2019.2.0" support for user-specific services were added.
    /Users/foo/Library/LaunchAgents/

.. note::
    As of the 2019.2.0 release, if a service is located in a ``LaunchAgent``
    path and a ``runas`` user is NOT specified, the current console user will
    be used to properly interact with the service.

"""
from __future__ import absolute_import, print_function, unicode_literals

# Import python libs
import logging
import os
import re

# Import salt libs
import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.stringutils
from salt.exceptions import CommandExecutionError

# Import 3rd party libs
from salt.ext import six
from salt.utils.versions import LooseVersion as _LooseVersion

# Define the module's virtual name
__virtualname__ = "service"

__func_alias__ = {
    "list_": "list",
}

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only for macOS with launchctl
    """
    if not salt.utils.platform.is_darwin():
        return (
            False,
            "Failed to load the mac_service module:\n"
            "Only available on macOS systems.",
        )

    if not salt.utils.path.which("launchctl"):
        return (
            False,
            "Failed to load the mac_service module:\n"
            'Required binary not found: "launchctl"',
        )

    if not salt.utils.path.which("plutil"):
        return (
            False,
            "Failed to load the mac_service module:\n"
            'Required binary not found: "plutil"',
        )

    if _LooseVersion(__grains__["osrelease"]) < _LooseVersion("10.11"):
        return (
            False,
            "Failed to load the mac_service module:\nRequires macOS 10.11 or newer",
        )

    return __virtualname__


def _name_in_services(name, services):
    """
    Checks to see if the given service is in the given services.

    :param str name: Service label, file name, or full path

    :param dict services: The currently available services.

    :return: The service information for the service, otherwise
    an empty dictionary

    :rtype: dict
    """
    if name in services:
        # Match on label
        return services[name]

    for service in six.itervalues(services):
        if service["file_path"].lower() == name:
            # Match on full path
            return service
        basename, ext = os.path.splitext(service["file_name"])
        if basename.lower() == name:
            # Match on basename
            return service

    return dict()


def _get_service(name):
    """
    Get information about a service.  If the service is not found, raise an
    error

    :param str name: Service label, file name, or full path

    :return: The service information for the service, otherwise an Error
    :rtype: dict
    """
    services = __utils__["mac_utils.available_services"]()
    name = name.lower()

    service = _name_in_services(name, services)

    # if we would the service we can return it
    if service:
        return service

    # if we got here our service is not available, now we can check to see if
    # we received a cached batch of services, if not we did a fresh check
    # so we need to raise that the service could not be found.
    try:
        if not __context__["using_cached_services"]:
            raise CommandExecutionError("Service not found: {0}".format(name))
    except KeyError:
        pass

    # we used a cached version to check, a service could have been made
    # between now and then, we should refresh our available services.
    services = __utils__["mac_utils.available_services"](refresh=True)

    # check to see if we found the service we are looking for.
    service = _name_in_services(name, services)

    if not service:
        # Could not find the service after refresh raise.
        raise CommandExecutionError("Service not found: {0}".format(name))

    # found it :)
    return service


def _always_running_service(name):
    """
    Check if the service should always be running based on the KeepAlive Key
    in the service plist.

    :param str name: Service label, file name, or full path

    :return: True if the KeepAlive key is set to True, False if set to False or
        not set in the plist at all.

    :rtype: bool

    .. versionadded:: 2019.2.0
    """

    # get all the info from the launchctl service
    service_info = show(name)

    # get the value for the KeepAlive key in service plist
    try:
        keep_alive = service_info["plist"]["KeepAlive"]
    except KeyError:
        return False

    # check if KeepAlive is True and not just set.

    if isinstance(keep_alive, dict):
        # check for pathstate
        for _file, value in six.iteritems(keep_alive.get("PathState", {})):
            if value is True and os.path.exists(_file):
                return True
            elif value is False and not os.path.exists(_file):
                return True

    if keep_alive is True:
        return True

    return False


def _get_domain_target(name, service_target=False):
    """
    Returns the domain/service target and path for a service. This is used to
    determine whether or not a service should be loaded in a user space or
    system space.

    :param str name: Service label, file name, or full path

    :param bool service_target: Whether to return a full
    service target. This is needed for the enable and disable
    subcommands of /bin/launchctl. Defaults to False

    :return: Tuple of the domain/service target and the path to the service.

    :rtype: tuple

    .. versionadded:: 2019.2.0
    """

    # Get service information
    service = _get_service(name)

    # get the path to the service
    path = service["file_path"]

    # most of the time we'll be at the system level.
    domain_target = "system"

    # check if a LaunchAgent as we should treat these differently.
    if "LaunchAgents" in path:
        # Get the console user so we can service in the correct session
        uid = __utils__["mac_utils.console_user"]()
        domain_target = "gui/{}".format(uid)

    # check to see if we need to make it a full service target.
    if service_target is True:
        domain_target = "{}/{}".format(domain_target, service["plist"]["Label"])

    return (domain_target, path)


def _launch_agent(name):
    """
    Checks to see if the provided service is a LaunchAgent

    :param str name: Service label, file name, or full path

    :return: True if a LaunchAgent, False if not.

    :rtype: bool

    .. versionadded:: 2019.2.0
    """

    # Get the path to the service.
    path = _get_service(name)["file_path"]

    if "LaunchAgents" not in path:
        return False
    return True


def show(name):
    """
    Show properties of a launchctl service

    :param str name: Service label, file name, or full path

    :return: The service information if the service is found
    :rtype: dict

    CLI Example:

    .. code-block:: bash

        salt '*' service.show org.cups.cupsd  # service label
        salt '*' service.show org.cups.cupsd.plist  # file name
        salt '*' service.show /System/Library/LaunchDaemons/org.cups.cupsd.plist  # full path
    """
    return _get_service(name)


def launchctl(sub_cmd, *args, **kwargs):
    """
    Run a launchctl command and raise an error if it fails

    :param str sub_cmd: Sub command supplied to launchctl

    :param tuple args: Tuple containing additional arguments to pass to
        launchctl

    :param dict kwargs: Dictionary containing arguments to pass to
        ``cmd.run_all``

    :param bool return_stdout: A keyword argument.  If true return the stdout
        of the launchctl command

    :return: ``True`` if successful, raise ``CommandExecutionError`` if not, or
        the stdout of the launchctl command if requested
    :rtype: bool, str

    CLI Example:

    .. code-block:: bash

        salt '*' service.launchctl debug org.cups.cupsd
    """
    return __utils__["mac_utils.launchctl"](sub_cmd, *args, **kwargs)


def list_(name=None, runas=None):
    """
    Run launchctl list and return the output

    :param str name: The name of the service to list

    :param str runas: User to run launchctl commands

    :return: If a name is passed returns information about the named service,
        otherwise returns a list of all services and pids
    :rtype: str

    CLI Example:

    .. code-block:: bash

        salt '*' service.list
        salt '*' service.list org.cups.cupsd
    """
    if name:
        # Get service information and label
        service = _get_service(name)
        label = service["plist"]["Label"]

        # we can assume if we are trying to list a LaunchAgent we need
        # to run as a user, if not provided, we'll use the console user.
        if not runas and _launch_agent(name):
            runas = __utils__["mac_utils.console_user"](username=True)

        # Collect information on service: will raise an error if it fails
        return launchctl("list", label, return_stdout=True, runas=runas)

    # Collect information on all services: will raise an error if it fails
    return launchctl("list", return_stdout=True, runas=runas)


def enable(name, runas=None):
    """
    Enable a launchd service. Raises an error if the service fails to be enabled

    :param str name: Service label, file name, or full path

    :param str runas: User to run launchctl commands

    :return: ``True`` if successful or if the service is already enabled
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' service.enable org.cups.cupsd
    """
    # Get the domain target. enable requires a full <service-target>
    service_target = _get_domain_target(name, service_target=True)[0]

    # Enable the service: will raise an error if it fails
    return launchctl("enable", service_target, runas=runas)


def disable(name, runas=None):
    """
    Disable a launchd service. Raises an error if the service fails to be
    disabled

    :param str name: Service label, file name, or full path

    :param str runas: User to run launchctl commands

    :return: ``True`` if successful or if the service is already disabled
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' service.disable org.cups.cupsd
    """
    # Get the service target. enable requires a full <service-target>
    service_target = _get_domain_target(name, service_target=True)[0]

    # disable the service: will raise an error if it fails
    return launchctl("disable", service_target, runas=runas)


def start(name, runas=None):
    """
    Start a launchd service.  Raises an error if the service fails to start

    .. note::
        To start a service in macOS the service must be enabled first. Use
        ``service.enable`` to enable the service.

    :param str name: Service label, file name, or full path

    :param str runas: User to run launchctl commands

    :return: ``True`` if successful or if the service is already running
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' service.start org.cups.cupsd
    """
    # Get the domain target.
    domain_target, path = _get_domain_target(name)

    # Load (bootstrap) the service: will raise an error if it fails
    return launchctl("bootstrap", domain_target, path, runas=runas)


def stop(name, runas=None):
    """
    Stop a launchd service.  Raises an error if the service fails to stop

    .. note::
        Though ``service.stop`` will unload a service in macOS, the service
        will start on next boot unless it is disabled. Use ``service.disable``
        to disable the service

    :param str name: Service label, file name, or full path

    :param str runas: User to run launchctl commands

    :return: ``True`` if successful or if the service is already stopped
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' service.stop org.cups.cupsd
    """
    # Get the domain target.
    domain_target, path = _get_domain_target(name)

    # Stop (bootout) the service: will raise an error if it fails
    return launchctl("bootout", domain_target, path, runas=runas)


def restart(name, runas=None):
    """
    Unloads and reloads a launchd service.  Raises an error if the service
    fails to reload

    :param str name: Service label, file name, or full path

    :param str runas: User to run launchctl commands

    :return: ``True`` if successful
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' service.restart org.cups.cupsd
    """
    # Restart the service: will raise an error if it fails
    if enabled(name):
        stop(name, runas=runas)
    start(name, runas=runas)

    return True


def status(name, sig=None, runas=None):
    """
    Return the status for a service.

    :param str name: Used to find the service from launchctl.  Can be any part
        of the service name or a regex expression.

    :param str sig: Find the service with status.pid instead.  Note that
        ``name`` must still be provided.

    :param str runas: User to run launchctl commands

    :return: The PID for the service if it is running, or 'loaded' if the
        service should not always have a PID, or otherwise an empty string

    :rtype: str

    CLI Example:

    .. code-block:: bash

        salt '*' service.status cups
    """
    # Find service with ps
    if sig:
        return __salt__["status.pid"](sig)

    try:
        _get_service(name)
    except CommandExecutionError as msg:
        log.error(msg)
        return ""

    if not runas and _launch_agent(name):
        runas = __utils__["mac_utils.console_user"](username=True)

    output = list_(runas=runas)

    # Used a string here instead of a list because that's what the linux version
    # of this module does
    pids = ""
    for line in output.splitlines():
        if "PID" in line:
            continue
        if re.search(name, line.split()[-1]):
            if line.split()[0].isdigit():
                if pids:
                    pids += "\n"
                pids += line.split()[0]

    # mac services are a little different than other platforms as they may be
    # set to run on intervals and may not always active with a PID. This will
    # return a string 'loaded' if it shouldn't always be running and is enabled.
    if not _always_running_service(name) and enabled(name) and not pids:
        return "loaded"

    return pids


def available(name):
    """
    Check that the given service is available.

    :param str name: The name of the service

    :return: True if the service is available, otherwise False
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' service.available com.openssh.sshd
    """
    try:
        _get_service(name)
        return True
    except CommandExecutionError:
        return False


def missing(name):
    """
    The inverse of service.available
    Check that the given service is not available.

    :param str name: The name of the service

    :return: True if the service is not available, otherwise False
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' service.missing com.openssh.sshd
    """
    return not available(name)


def enabled(name, runas=None):
    """
    Check if the specified service is enabled

    :param str name: The name of the service to look up

    :param str runas: User to run launchctl commands

    :return: True if the specified service enabled, otherwise False
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' service.enabled org.cups.cupsd
    """
    # Try to list the service.  If it can't be listed, it's not enabled
    try:
        list_(name=name, runas=runas)
        return True
    except CommandExecutionError:
        return False


def disabled(name, runas=None, domain="system"):
    """
    Check if the specified service is not enabled. This is the opposite of
    ``service.enabled``

    :param str name: The name to look up

    :param str runas: User to run launchctl commands

    :param str domain: domain to check for disabled services. Default is system.

    :return: True if the specified service is NOT enabled, otherwise False
    :rtype: bool

    CLI Example:

    .. code-block:: bash

        salt '*' service.disabled org.cups.cupsd
    """

    disabled = launchctl("print-disabled", domain, return_stdout=True, runas=runas)
    for service in disabled.split("\n"):
        if name in service:
            srv_name = service.split("=>")[0].split('"')[1]
            status = service.split("=>")[1]
            if name != srv_name:
                pass
            else:
                return True if "true" in status.lower() else False

    return False


def get_all(runas=None):
    """
    Return a list of services that are enabled or available. Can be used to
    find the name of a service.

    :param str runas: User to run launchctl commands

    :return: A list of all the services available or enabled
    :rtype: list

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_all
    """
    # Get list of enabled services
    enabled = get_enabled(runas=runas)

    # Get list of all services
    available = list(__utils__["mac_utils.available_services"]().keys())

    # Return composite list
    return sorted(set(enabled + available))


def get_enabled(runas=None):
    """
    Return a list of all services that are enabled. Can be used to find the
    name of a service.

    :param str runas: User to run launchctl commands

    :return: A list of all the services enabled on the system
    :rtype: list

    CLI Example:

    .. code-block:: bash

        salt '*' service.get_enabled
    """
    # Collect list of enabled services
    stdout = list_(runas=runas)
    service_lines = [line for line in stdout.splitlines()]

    # Construct list of enabled services
    enabled = []
    for line in service_lines:
        # Skip header line
        if line.startswith("PID"):
            continue

        pid, status, label = line.split("\t")
        enabled.append(label)

    return sorted(set(enabled))



# -*- coding: utf-8 -*-
"""
Unit tests for salt.utils.color.py
"""

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

# Import Salt libs
import salt.utils.color
from salt.ext import six

# Import Salt Testing libs
from tests.support.unit import TestCase


class ColorUtilsTestCase(TestCase):
    def test_get_colors(self):
        ret = salt.utils.color.get_colors()
        self.assertEqual("\x1b[0;37m", six.text_type(ret["LIGHT_GRAY"]))

        ret = salt.utils.color.get_colors(use=False)
        self.assertDictContainsSubset({"LIGHT_GRAY": ""}, ret)

        ret = salt.utils.color.get_colors(use="LIGHT_GRAY")
        # LIGHT_YELLOW now == LIGHT_GRAY
        self.assertEqual(
            six.text_type(ret["LIGHT_YELLOW"]), six.text_type(ret["LIGHT_GRAY"])
        )



# -*- coding: utf-8 -*-
"""
    tests.unit.utils.cache_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Test the salt cache objects
"""

# Import python libs
from __future__ import absolute_import, print_function, unicode_literals

import os
import shutil
import tempfile
import time

# Import salt libs
import salt.config
import salt.loader
import salt.payload
import salt.utils.cache as cache
import salt.utils.data
import salt.utils.files

# Import Salt Testing libs
from tests.support.unit import TestCase


class CacheDictTestCase(TestCase):
    def test_sanity(self):
        """
        Make sure you can instantiate etc.
        """
        cd = cache.CacheDict(5)
        self.assertIsInstance(cd, cache.CacheDict)

        # do some tests to make sure it looks like a dict
        self.assertNotIn("foo", cd)
        cd["foo"] = "bar"
        self.assertEqual(cd["foo"], "bar")
        del cd["foo"]
        self.assertNotIn("foo", cd)

    def test_ttl(self):
        cd = cache.CacheDict(0.1)
        cd["foo"] = "bar"
        self.assertIn("foo", cd)
        self.assertEqual(cd["foo"], "bar")
        time.sleep(0.2)
        self.assertNotIn("foo", cd)

        # make sure that a get would get a regular old key error
        self.assertRaises(KeyError, cd.__getitem__, "foo")


class CacheContextTestCase(TestCase):
    def setUp(self):
        context_dir = os.path.join(tempfile.gettempdir(), "context")
        if os.path.exists(context_dir):
            shutil.rmtree(os.path.join(tempfile.gettempdir(), "context"))

    def test_smoke_context(self):
        """
        Smoke test the context cache
        """
        if os.path.exists(os.path.join(tempfile.gettempdir(), "context")):
            self.skipTest("Context dir already exists")
        else:
            opts = salt.config.DEFAULT_MINION_OPTS.copy()
            opts["cachedir"] = tempfile.gettempdir()
            context_cache = cache.ContextCache(opts, "cache_test")

            context_cache.cache_context({"a": "b"})

            ret = context_cache.get_cache_context()

            self.assertDictEqual({"a": "b"}, ret)

    def test_context_wrapper(self):
        """
        Test to ensure that a module which decorates itself
        with a context cache can store and retrieve its contextual
        data
        """
        opts = salt.config.DEFAULT_MINION_OPTS.copy()
        opts["cachedir"] = tempfile.gettempdir()

        ll_ = salt.loader.LazyLoader(
            [os.path.join(os.path.dirname(os.path.realpath(__file__)), "cache_mods")],
            tag="rawmodule",
            virtual_enable=False,
            opts=opts,
        )

        cache_test_func = ll_["cache_mod.test_context_module"]

        self.assertEqual(cache_test_func()["called"], 0)
        self.assertEqual(cache_test_func()["called"], 1)


__context__ = {"a": "b"}
__opts__ = {"cachedir": "/tmp"}


class ContextCacheTest(TestCase):
    """
    Test case for salt.utils.cache.ContextCache
    """

    def setUp(self):
        """
        Clear the cache before every test
        """
        context_dir = os.path.join(__opts__["cachedir"], "context")
        if os.path.isdir(context_dir):
            shutil.rmtree(context_dir)

    def test_set_cache(self):
        """
        Tests to ensure the cache is written correctly
        """

        @cache.context_cache
        def _test_set_cache():
            """
            This will inherit globals from the test module itself.
            Normally these are injected by the salt loader [salt.loader]
            """

        _test_set_cache()

        target_cache_file = os.path.join(
            __opts__["cachedir"], "context", "{0}.p".format(__name__)
        )
        self.assertTrue(
            os.path.isfile(target_cache_file), "Context cache did not write cache file"
        )

        # Test manual de-serialize
        with salt.utils.files.fopen(target_cache_file, "rb") as fp_:
            target_cache_data = salt.utils.data.decode(
                salt.payload.Serial(__opts__).load(fp_)
            )
        self.assertDictEqual(__context__, target_cache_data)

        # Test cache de-serialize
        cc = cache.ContextCache(__opts__, __name__)
        retrieved_cache = cc.get_cache_context()
        self.assertDictEqual(retrieved_cache, __context__)

    def test_refill_cache(self):
        """
        Tests to ensure that the context cache can rehydrate a wrapped function
        """
        # First populate the cache
        @cache.context_cache
        def _test_set_cache():
            pass

        _test_set_cache()

        # Then try to rehydate a func
        @cache.context_cache
        def _test_refill_cache(comparison_context):
            self.assertEqual(__context__, comparison_context)

        global __context__
        __context__ = {}
        _test_refill_cache({"a": "b"})  # Compare to the context before it was emptied


class CacheDiskTestCase(TestCase):
    def test_everything(self):
        """
        Make sure you can instantiate, add, update, remove, expire
        """
        try:
            tmpdir = tempfile.mkdtemp()
            path = os.path.join(tmpdir, "CacheDisk_test")

            # test instantiation
            cd = cache.CacheDisk(0.1, path)
            self.assertIsInstance(cd, cache.CacheDisk)

            # test to make sure it looks like a dict
            self.assertNotIn("foo", cd)
            cd["foo"] = "bar"
            self.assertIn("foo", cd)
            self.assertEqual(cd["foo"], "bar")
            del cd["foo"]
            self.assertNotIn("foo", cd)

            # test persistence
            cd["foo"] = "bar"
            cd2 = cache.CacheDisk(0.1, path)
            self.assertIn("foo", cd2)
            self.assertEqual(cd2["foo"], "bar")

            # test ttl
            time.sleep(0.2)
            self.assertNotIn("foo", cd)
            self.assertNotIn("foo", cd2)

        finally:
            shutil.rmtree(tmpdir, ignore_errors=True)
